from __future__ import annotations

import io
from abc import ABC, abstractmethod
from enum import IntEnum
from typing import Generic, TypeVar

from dissect.target.helpers.sunrpc import sunrpc

ProcedureParams = TypeVar("ProcedureParams")
ProcedureResults = TypeVar("ProcedureResults")
Credentials = TypeVar("Credentials")
Verifier = TypeVar("Verifier")
Serializable = TypeVar("Serializable")
AuthProtocol = TypeVar("AuthProtocol")
EnumType = TypeVar("EnumType", bound=IntEnum)
ElementType = TypeVar("ElementType")


class MessageType(IntEnum):
    CALL = 0
    REPLY = 1


ALIGNMENT = 4


class Serializer(ABC, Generic[Serializable]):
    @abstractmethod
    def serialize(self, _: Serializable) -> bytes:
        pass


class Deserializer(ABC, Generic[Serializable]):
    def deserialize_from_bytes(self, payload: bytes) -> Serializable:
        return self.deserialize(io.BytesIO(payload))

    @abstractmethod
    def deserialize(self, _: io.BytesIO) -> Serializable:
        pass


# Unfortunately xdrlib is deprecated in Python 3.11, so we implement the following serialization classes
# See https://datatracker.ietf.org/doc/html/rfc1014 for the XDR specification.


class Int32Serializer(Serializer[int], Deserializer[int]):
    _signed = True

    def serialize(self, i: int) -> bytes:
        return i.to_bytes(length=4, byteorder="big", signed=self._signed)

    def deserialize(self, payload: io.BytesIO) -> int:
        return int.from_bytes(payload.read(4), byteorder="big", signed=self._signed)


class UInt32Serializer(Int32Serializer):
    _signed = False


class OpaqueVarLengthSerializer(Serializer[bytes], Deserializer[bytes]):
    def serialize(self, body: bytes) -> bytes:
        length = len(body)
        result = UInt32Serializer().serialize(length)
        result += body

        padding_bytes = (ALIGNMENT - (length % ALIGNMENT)) % ALIGNMENT
        return result + b"\x00" * padding_bytes

    def deserialize(self, payload: io.BytesIO) -> bytes:
        length = UInt32Serializer().deserialize(payload)
        result = payload.read(length)
        padding_bytes = (ALIGNMENT - (length % ALIGNMENT)) % ALIGNMENT
        payload.read(padding_bytes)
        return result


class StringSerializer(Serializer[str], Deserializer[str]):
    def serialize(self, s: str) -> bytes:
        return OpaqueVarLengthSerializer().serialize(s.encode("ascii"))

    def deserialize(self, payload: io.BytesIO) -> str:
        return OpaqueVarLengthSerializer().deserialize(payload).decode("ascii")


class XdrSerializer(Serializer[Serializable], Generic[Serializable]):
    def _write_uint32(self, i: int) -> bytes:
        return UInt32Serializer().serialize(i)

    def _write_int32(self, i: int) -> bytes:
        return Int32Serializer().serialize(i)

    def _write_uint64(self, i: int) -> bytes:
        return i.to_bytes(length=8, byteorder="big", signed=False)

    def _write_enum(self, enum: EnumType) -> bytes:
        return self._write_int32(enum.value)

    def _write_var_length(self, elements: list[ElementType], serializer: Serializer[ElementType]) -> bytes:
        result = self._write_uint32(len(elements))
        payload = (serializer.serialize(element) for element in elements)
        return result + b"".join(payload)

    def _write_var_length_opaque(self, body: bytes) -> bytes:
        return OpaqueVarLengthSerializer().serialize(body)

    def _write_string(self, s: str) -> bytes:
        return StringSerializer().serialize(s)


class XdrDeserializer(Deserializer[Serializable], Generic[Serializable]):
    def _read_uint32(self, payload: io.BytesIO) -> int:
        return UInt32Serializer().deserialize(payload)

    def _read_int32(self, payload: io.BytesIO) -> int:
        return Int32Serializer().deserialize(payload)

    def _read_uint64(self, payload: io.BytesIO) -> int:
        return int.from_bytes(payload.read(8), byteorder="big", signed=False)

    def _read_enum(self, payload: io.BytesIO, enum: type[EnumType]) -> EnumType:
        value = self._read_int32(payload)
        return enum(value)

    def _read_var_length(self, payload: io.BytesIO, deserializer: Deserializer[ElementType]) -> list[ElementType]:
        length = self._read_uint32(payload)
        return [deserializer.deserialize(payload) for _ in range(length)]

    def _read_var_length_opaque(self, payload: io.BytesIO) -> bytes:
        return OpaqueVarLengthSerializer().deserialize(payload)

    def _read_string(self, payload: io.BytesIO) -> str:
        return StringSerializer().deserialize(payload)

    def _read_optional(self, payload: io.BytesIO, deserializer: Deserializer[ElementType]) -> ElementType | None:
        has_value = self._read_enum(payload, sunrpc.Bool)
        if has_value == sunrpc.Bool.FALSE:
            return None
        return deserializer.deserialize(payload)


class ReplyStat(IntEnum):
    MSG_ACCEPTED = 0
    MSG_DENIED = 1


class AuthFlavor(IntEnum):
    AUTH_NULL = 0
    AUTH_UNIX = 1
    AUTH_SHORT = 2
    AUTH_DES = 3


class AuthSerializer(XdrSerializer[AuthProtocol], XdrDeserializer[AuthProtocol], Generic[AuthProtocol]):
    def serialize(self, protocol: AuthProtocol) -> bytes:
        flavor = self._flavor()
        result = self._write_int32(flavor)

        body = self._write_body(protocol)
        return result + self._write_var_length_opaque(body)

    def deserialize(self, payload: io.BytesIO) -> AuthProtocol:
        flavor = self._read_int32(payload)
        if flavor != self._flavor():
            raise ValueError(f"Expected flavor {self._flavor()}, got {flavor}")

        body = self._read_var_length_opaque(payload)
        return self._read_body(io.BytesIO(body))

    # The return type is not AuthFlavor, because AuthProtocol is open to extension.
    @abstractmethod
    def _flavor(self) -> int:
        pass

    @abstractmethod
    def _write_body(self, _: AuthProtocol) -> bytes:
        pass

    @abstractmethod
    def _read_body(self, _: io.BytesIO) -> AuthProtocol:
        pass


class AuthNullSerializer(AuthSerializer[sunrpc.AuthNull]):
    def _flavor(self) -> int:
        return AuthFlavor.AUTH_NULL.value

    def _write_body(self, _: AuthProtocol) -> bytes:
        return b""

    def _read_body(self, _: io.BytesIO) -> AuthProtocol:
        return sunrpc.AuthNull()


class AuthUnixSerializer(AuthSerializer[sunrpc.AuthUnix]):
    def _flavor(self) -> int:
        return AuthFlavor.AUTH_UNIX.value

    def _write_body(self, protocol: sunrpc.AuthUnix) -> bytes:
        result = self._write_uint32(protocol.stamp)
        result += self._write_string(protocol.machinename)
        result += self._write_uint32(protocol.uid)
        result += self._write_uint32(protocol.gid)
        result += self._write_var_length(protocol.gids, Int32Serializer())
        return result

    def _read_body(self, payload: io.BytesIO) -> sunrpc.AuthUnix:
        stamp = self._read_uint32(payload)
        machinename = self._read_string(payload)
        uid = self._read_uint32(payload)
        gid = self._read_uint32(payload)
        gids = self._read_var_length(payload, Int32Serializer())
        return sunrpc.AuthUnix(stamp, machinename, uid, gid, gids)


class MessageSerializer(
    XdrSerializer[sunrpc.Message[ProcedureParams, ProcedureResults, Credentials, Verifier]],
    XdrDeserializer[sunrpc.Message[ProcedureParams, ProcedureResults, Credentials, Verifier]],
    Generic[ProcedureParams, ProcedureResults, Credentials, Verifier],
):
    def __init__(
        self,
        params_serializer: XdrSerializer[ProcedureParams],
        results_deserializer: XdrDeserializer[ProcedureResults],
        credentials_serializer: AuthSerializer[Credentials],
        verifier_serializer: AuthSerializer[Verifier],
    ):
        self._params_serializer = params_serializer
        self._results_deserializer = results_deserializer
        self._credentials_serializer = credentials_serializer
        self._verifier_serializer = verifier_serializer

    def serialize(self, message: sunrpc.Message[ProcedureParams, ProcedureResults, Credentials, Verifier]) -> bytes:
        if not isinstance(message.body, sunrpc.CallBody):
            raise NotImplementedError("Only CALL messages are serializable")

        result = self._write_uint32(message.xid)
        result += self._write_enum(MessageType.CALL)
        return result + self._write_call_body(message.body)

    def deserialize(
        self, payload: io.BytesIO
    ) -> sunrpc.Message[ProcedureParams, ProcedureResults, Credentials, Verifier]:
        xid = self._read_uint32(payload)
        message_type = self._read_enum(payload, MessageType)
        if message_type != MessageType.REPLY:
            raise NotImplementedError("Only REPLY messages are deserializable")

        reply_stat = self._read_enum(payload, ReplyStat)
        if reply_stat == ReplyStat.MSG_ACCEPTED:
            reply = self._read_accepted_reply(payload)
        elif reply_stat == ReplyStat.MSG_DENIED:
            reply = self._read_rejected_reply(payload)
        else:
            # assert_never is slightly better, but only available starting from Python 3.11
            raise ValueError(f"Unexpected value for reply_stat {reply_stat}")

        return sunrpc.Message(xid, reply)

    def _write_call_body(self, call_body: sunrpc.CallBody) -> bytes:
        result = self._write_uint32(call_body.rpc_version)
        result += self._write_uint32(call_body.program)
        result += self._write_uint32(call_body.version)
        result += self._write_uint32(call_body.procedure)
        result += self._credentials_serializer.serialize(call_body.cred)
        result += self._verifier_serializer.serialize(call_body.verf)
        result += self._params_serializer.serialize(call_body.params)
        return result

    def _read_accepted_reply(self, payload: io.BytesIO) -> sunrpc.AcceptedReply[ProcedureResults, Verifier]:
        verf = self._verifier_serializer.deserialize(payload)
        stat = self._read_enum(payload, sunrpc.AcceptStat)
        if stat == sunrpc.AcceptStat.SUCCESS:
            results = self._results_deserializer.deserialize(payload)
        elif stat == sunrpc.AcceptStat.PROG_MISMATCH:
            results = self._read_mismatch(payload)
        else:
            # Void in case of PROG_UNAVAIL, PROC_UNAVAIL, GARBAGE_ARGS
            results = None

        return sunrpc.AcceptedReply(verf, stat, results)

    def _read_rejected_reply(self, payload: io.BytesIO) -> sunrpc.RejectedReply:
        reject_stat = self._read_enum(payload, sunrpc.RejectStat)
        if reject_stat == sunrpc.RejectStat.RPC_MISMATCH:
            mismatch = self._read_mismatch(payload)
            return sunrpc.RejectedReply(reject_stat, mismatch)
        if reject_stat == sunrpc.RejectStat.AUTH_ERROR:
            auth_stat = self._read_enum(payload, sunrpc.AuthStat)
            return sunrpc.RejectedReply(reject_stat, auth_stat)

        # assert_never is slightly better, but only available starting from Python 3.11
        raise ValueError(f"Unexpected value for reject_stat: {reject_stat}")

    def _read_mismatch(self, payload: io.BytesIO) -> sunrpc.Mismatch:
        low = self._read_uint32(payload)
        high = self._read_uint32(payload)
        return sunrpc.Mismatch(low, high)


class PortMappingSerializer(XdrSerializer[sunrpc.PortMapping]):
    def serialize(self, port_mapping: sunrpc.PortMapping) -> bytes:
        result = self._write_uint32(port_mapping.program)
        result += self._write_uint32(port_mapping.version)
        result += self._write_enum(port_mapping.protocol)
        result += self._write_uint32(port_mapping.port)
        return result



from __future__ import annotations

from typing import TYPE_CHECKING

from dissect.target.filesystems.dir import DirectoryFilesystem
from dissect.target.helpers.record import WindowsUserRecord
from dissect.target.loader import Loader
from dissect.target.plugin import OSPlugin, export
from dissect.target.plugins.os.windows.registry import RegistryPlugin

if TYPE_CHECKING:
    from collections.abc import Iterator
    from pathlib import Path

    from typing_extensions import Self

    from dissect.target.filesystem import Filesystem
    from dissect.target.target import Target


class ProfileLoader(Loader):
    """Load NTUSER.DAT files."""

    @staticmethod
    def detect(path: Path) -> bool:
        return bool(path.is_dir() and path.joinpath("NTUSER.DAT").exists())

    def map(self, target: Target) -> None:
        username = self.path.name

        dfs = DirectoryFilesystem(self.absolute_path, case_sensitive=False)
        target.filesystems.add(dfs)
        target.fs.mount(f"sysvol/users/{username}", dfs)

        target._os_plugin = ProfileOSPlugin
        target.add_plugin(RegistryPlugin, check_compatible=False)


class ProfileOSPlugin(OSPlugin):
    @classmethod
    def detect(cls, target: Target) -> bool:
        return True

    @classmethod
    def create(cls, target: Target, sysvol: Filesystem) -> Self:
        return cls(target)

    @export(property=True)
    def hostname(self) -> str:
        return self.target._generic_name

    @export(property=True)
    def ips(self) -> list:
        return []

    @export(property=True)
    def version(self) -> None:
        return None

    @export
    def users(self) -> Iterator[WindowsUserRecord]:
        yield WindowsUserRecord(
            sid="0",
            name=self.hostname,
            home=f"sysvol/users/{self.hostname}",
            _target=self.target,
        )

    @export(property=True)
    def os(self) -> str:
        return "windows"



from __future__ import annotations

from datetime import datetime, timezone
from typing import TYPE_CHECKING

import pytest

from dissect.target.plugins.apps.virtualization.vmware_workstation import (
    VmwareWorkstationPlugin,
)
from tests._utils import absolute_path

if TYPE_CHECKING:
    from dissect.target.filesystem import VirtualFilesystem
    from dissect.target.target import Target


@pytest.mark.parametrize(
    ("target", "fs", "dnd_path"),
    [
        pytest.param("target_win_users", "fs_win", "Users\\John\\AppData\\Local\\Temp\\VmwareDND", id="windows"),
        pytest.param("target_unix_users", "fs_unix", "/home/user/.cache/vmware/drag_and_drop", id="linux"),
    ],
)
def test_vmware_workstation_clipboard_dnd(
    target: Target, fs: VirtualFilesystem, dnd_path: str, request: pytest.FixtureRequest
) -> None:
    """Test if we correctly detect drag and drop artifacts on Windows and Unix targets."""

    target = request.getfixturevalue(target)
    fs = request.getfixturevalue(fs)
    fs.map_dir_from_tar(dnd_path, absolute_path("_data/plugins/apps/virtualization/vmware_workstation/dnd.tar"))
    target.add_plugin(VmwareWorkstationPlugin)
    results = sorted(target.vmware.clipboard(), key=lambda r: r.path)
    dnd_path = dnd_path.replace("\\", "/")

    assert len(results) == 3
    assert [str(r.path).replace("C:\\", "").replace("\\", "/") for r in results] == [
        f"{dnd_path}/8lkBf1/wachter.jpg",
        f"{dnd_path}/iUpQHG/credentials.txt",
        f"{dnd_path}/x1G9fJ/example.txt",
    ]
    assert results[0].ts == datetime(2024, 9, 18, 11, 16, 44, tzinfo=timezone.utc)


@pytest.mark.parametrize(
    ("target", "fs", "inventory_dir"),
    [
        pytest.param(
            "target_win_users",
            "fs_win",
            "Users\\John\\AppData\\Roaming\\VMware\\inventory.vmls",
            id="windows",
        ),
        pytest.param(
            "target_unix_users",
            "fs_unix",
            "/home/user/.vmware/inventory.vmls",
            id="linux",
        ),
    ],
)
def test_vmware_workstation_vm_configs(
    target: Target, fs: VirtualFilesystem, inventory_dir: str, request: pytest.FixtureRequest
) -> None:
    """Test if we find and parse VM configuration files correctly."""

    target = request.getfixturevalue(target)
    fs = request.getfixturevalue(fs)
    fs.map_file(
        inventory_dir,
        str(absolute_path(f"_data/plugins/apps/virtualization/vmware_workstation/inventory-{target.os}.vmls")),
    )
    fs.map_file(
        "first.vmx" if target.os == "windows" else "/path/to/first.vmx",
        str(absolute_path("_data/plugins/apps/virtualization/vmware_workstation/first.vmx")),
    )

    target.add_plugin(VmwareWorkstationPlugin)

    records = list(target.vmware.config())

    assert len(records) == 3

    assert records[0].ts
    assert records[0].name == "First Virtual Machine"
    assert not records[0].is_clone
    assert not records[0].is_favorite
    assert records[0].state == "normal"
    assert records[0].uuid == "41cca13579fa4366abef2582d21a0c20"
    assert records[0].annotation == "This is an example annotation."
    assert records[0].mac_addresses == ["00:0C:29:12:34:56"]
    assert records[0].disks == ["First Virtual Machine.vmdk"]
    assert records[0].sources in (
        ["C:\\Users\\John\\AppData\\Roaming\\VMware\\inventory.vmls", "sysvol\\first.vmx"],
        ["/home/user/.vmware/inventory.vmls", "/path/to/first.vmx"],
    )



from __future__ import annotations

from io import BytesIO

from dissect.target.filesystem import VirtualFilesystem
from dissect.target.plugins.os.unix.esxi._os import ESXiPlugin
from dissect.target.plugins.os.unix.esxi.vm import VirtualMachinePlugin
from dissect.target.target import Target


def test_vm() -> None:
    """Test that the ESXi VM inventory plugin yields the correct records."""
    vfs = VirtualFilesystem()
    vfs.map_file_fh(
        "/etc/vmware/hostd/vmInventory.xml",
        BytesIO(b"""
          <ConfigRoot>
            <ConfigEntry id="0000">
              <objID>1</objID>
              <secDomain/>
              <vmxCfgPath>/vmfs/volumes/6800e48c-3dcffc58-6af7-bc2411ec8065/Alpine/Alpine.vmx</vmxCfgPath>
            </ConfigEntry>
          </ConfigRoot>
        """),
    )

    # vmx not defined in vmInventory
    vfs.map_file_fh(
        "/vmfs/volumes/6800e48c-3dcffc58-6af7-bc2411ec8065/Debian/Debian.vmx", BytesIO(b'displayName = "Debian"')
    )

    target = Target()
    target._os_plugin = ESXiPlugin(target)
    target.filesystems.add(vfs)
    target.fs.mount("/", vfs)
    target.apply()

    target.add_plugin(VirtualMachinePlugin)

    records = list(target.vm.inventory())
    assert len(records) == 1
    assert str(records[0].path) == "/vmfs/volumes/6800e48c-3dcffc58-6af7-bc2411ec8065/Alpine/Alpine.vmx"

    orphaned = list(target.vm.orphaned())
    assert len(orphaned) == 1
    assert str(orphaned[0].path) == "/vmfs/volumes/6800e48c-3dcffc58-6af7-bc2411ec8065/Debian/Debian.vmx"

    assert len(list(target.vm())) == 2



from __future__ import annotations

from typing import TYPE_CHECKING, BinaryIO

from dissect.volume import lvm

from dissect.target.helpers.logging import get_logger
from dissect.target.volume import LogicalVolumeSystem, Volume

if TYPE_CHECKING:
    from collections.abc import Iterator

    from typing_extensions import Self


log = get_logger(__name__)

OPEN_TYPES = (
    "linear",
    "striped",
    "mirror",
    "thin",
)

KNOWN_SKIP_TYPES = (
    "snapshot",
    "thin-pool",
)


class LvmVolumeSystem(LogicalVolumeSystem):
    __type__ = "lvm"

    def __init__(self, fh: BinaryIO | list[BinaryIO], *args, **kwargs):
        self.lvm = lvm.LVM2(fh)
        super().__init__(fh, *args, **kwargs)

    @classmethod
    def open_all(cls, volumes: list[BinaryIO]) -> Iterator[Self]:
        devices: dict[str, list[lvm.LVM2Device]] = {}

        for vol in volumes:
            if not cls.detect_volume(vol):
                continue

            dev = lvm.LVM2Device(vol)
            if metadata := dev.metadata:
                vg_name = next(key for key, value in metadata.items() if isinstance(value, dict))
                devices.setdefault(vg_name, []).append(dev)

        for pvs in devices.values():
            try:
                yield cls(pvs, disk=[pv.fh for pv in pvs])
            except Exception:  # noqa: PERF203
                continue

    @staticmethod
    def _detect(fh: BinaryIO) -> bool:
        vols = [fh] if not isinstance(fh, list) else fh
        return any(LvmVolumeSystem.detect_volume(vol) for vol in vols)

    @staticmethod
    def _detect_volume(fh: BinaryIO) -> bool:
        buf = fh.read(4096)
        return b"LABELONE" in buf

    def _volumes(self) -> Iterator[Volume]:
        num = 1

        for lv_name, lv in self.lvm.volume_group.logical_volumes.items():
            if lv.type not in OPEN_TYPES:
                if lv.type not in KNOWN_SKIP_TYPES:
                    log.debug("Skipping unsupported LVM logical volume type: %s (%s)", lv.type, lv)
                continue

            # When composing a vg-lv name, LVM2 replaces hyphens with double hyphens in the vg and lv names
            # Emulate that here for the volume name
            name = f"{lv.vg.name.replace('-', '--')}-{lv_name.replace('-', '--')}"

            fh = lv.open()
            yield Volume(fh, num, None, fh.size, None, name, raw=lv, disk=self.disk, vs=self)

            num += 1



from __future__ import annotations

import gzip
from typing import TYPE_CHECKING

import pytest

from dissect.target.filesystems.cramfs import CramfsFilesystem, CramfsFilesystemEntry, c_cramfs
from tests._utils import absolute_path

if TYPE_CHECKING:
    from collections.abc import Iterator


def test_cramfs_detect() -> None:
    """Test that CramFS filesystems are correctly detected."""
    with gzip.open(absolute_path("_data/filesystems/cramfs/cramfs.bin.gz"), "rb") as fh:
        assert CramfsFilesystem.detect(fh)
        assert CramfsFilesystem(fh).cramfs


@pytest.fixture
def cramfs() -> Iterator[CramfsFilesystemEntry]:
    with gzip.open(absolute_path("_data/filesystems/cramfs/cramfs.bin.gz"), "rb") as fh:
        cramfs = CramfsFilesystem(fh)
        yield cramfs


@pytest.mark.parametrize(
    ("entry_path", "expected_blocks"),
    [
        ("/bin", 1),
        ("/home", 0),
        ("/bin/dvrbox", 6),
    ],
)
def test_cramfs_stat(cramfs: CramfsFilesystem, entry_path: str, expected_blocks: int) -> None:
    """Test consistency in ``stat()`` results."""
    cramfs_entry = cramfs.get(entry_path)
    inode = cramfs_entry.entry

    stat = cramfs_entry.stat()
    assert stat.st_mode == inode.mode
    assert stat.st_ino == inode.offset
    assert stat.st_dev == id(cramfs_entry.fs)
    assert stat.st_nlink == 1
    assert stat.st_uid == inode.uid
    assert stat.st_gid == inode.gid
    assert stat.st_size == inode.size
    assert stat.st_blksize == c_cramfs.CRAMFS_BLOCK_SIZE
    assert stat.st_blocks == expected_blocks


def test_cramfs_symlinks(cramfs: CramfsFilesystem) -> None:
    symlink = cramfs.get("/bin/macGuarder")
    target = cramfs.get("/bin/dvrbox")

    assert symlink.is_file()
    assert symlink.is_symlink()
    assert symlink.name == "macGuarder"
    assert symlink.readlink() == "./dvrbox"
    assert symlink.readlink_ext().name == target.name

    assert target.is_file()
    assert target.name == "dvrbox"


def test_cramfs_dir(cramfs: CramfsFilesystem) -> None:
    dir = cramfs.get("/etc/init.d")
    assert dir.is_dir()

    entries = list(dir.iterdir())
    assert len(entries) == 2
    assert sorted(entries) == sorted(["dnode", "rcS"])

    entries = [entry.name for entry in dir.scandir()]
    assert dir.is_dir()
    assert len(entries) == 2
    assert sorted(entries) == sorted(["dnode", "rcS"])



from __future__ import annotations

import datetime
from io import BytesIO
from pathlib import Path
from typing import TYPE_CHECKING
from unittest.mock import Mock, patch

import pytest

from dissect.target.filesystem import VirtualFilesystem
from dissect.target.plugins.os.windows.amcache import AmcachePlugin
from dissect.target.plugins.os.windows.log.amcache import AmcacheInstallPlugin
from tests._utils import absolute_path

if TYPE_CHECKING:
    from collections.abc import Iterator

    from dissect.target.filesystem import VirtualFilesystem
    from dissect.target.target import Target


def test_amcache_new_format(target_win: Target, fs_win: VirtualFilesystem) -> None:
    amcache_file = absolute_path("_data/plugins/os/windows/amcache/amcache-new.hve")
    fs_win.map_file("windows/appcompat/programs/amcache.hve", amcache_file)

    target_win.add_plugin(AmcachePlugin)

    files = list(target_win.amcache.files())
    programs = list(target_win.amcache.programs())

    applications = list(target_win.amcache.applications())
    application_files = list(target_win.amcache.application_files())
    application_shortcuts = list(target_win.amcache.shortcuts())
    drivers = list(target_win.amcache.drivers())
    containers = list(target_win.amcache.device_containers())

    assert len(files) == 0
    assert len(programs) == 0
    assert len(applications) == 118
    assert len(application_files) == 222
    assert len(application_shortcuts) == 65
    assert len(containers) == 9
    assert len(drivers) == 361


def test_amcache_old_format(target_win: Target, fs_win: VirtualFilesystem) -> None:
    amcache_file = absolute_path("_data/plugins/os/windows/amcache/amcache-old.hve")
    fs_win.map_file("windows/appcompat/programs/amcache.hve", amcache_file)

    target_win.add_plugin(AmcachePlugin)

    files = list(target_win.amcache.files())
    programs = list(target_win.amcache.programs())

    applications = list(target_win.amcache.applications())
    application_files = list(target_win.amcache.application_files())
    application_shortcuts = list(target_win.amcache.shortcuts())
    drivers = list(target_win.amcache.drivers())
    containers = list(target_win.amcache.device_containers())

    assert len(files) == 69
    assert len(programs) == 0
    assert len(applications) == 326
    assert len(application_files) == 0
    assert len(application_shortcuts) == 0
    assert len(containers) == 16
    assert len(drivers) == 0


def test_amcache_windows_11_applaunches(target_win: Target, fs_win: VirtualFilesystem) -> None:
    # Test file taken from https://github.com/AndrewRathbun/DFIRArtifactMuseum/blob/main/Windows/Amcache/Win11/RathbunVM/PcaAppLaunchDic.txt
    # Licensed under the MIT License, Copyright (c) 2022 DFIR Artifact Museum
    applaunch_file = absolute_path("_data/plugins/os/windows/amcache/pca/PcaAppLaunchDic.txt")
    fs_win.map_file("windows/appcompat/pca/PcaAppLaunchDic.txt", applaunch_file)

    target_win.add_plugin(AmcachePlugin)
    applaunches = list(target_win.amcache.applaunches())

    assert len(applaunches) == 55
    assert applaunches[0].ts == datetime.datetime(2022, 12, 17, 13, 27, 53, 96000, tzinfo=datetime.timezone.utc)
    assert applaunches[0].path == "C:\\ProgramData\\Sophos\\AutoUpdate\\Cache\\sophos_autoupdate1.dir\\su-setup32.exe"


def test_amcache_windows_11_general(target_win: Target, fs_win: VirtualFilesystem) -> None:
    # Test files taken from https://github.com/AndrewRathbun/DFIRArtifactMuseum/blob/main/Windows/Amcache/Win11/RathbunVM/PcaGeneralDb0.txt
    # and https://github.com/AndrewRathbun/DFIRArtifactMuseum/blob/main/Windows/Amcache/Win11/RathbunVM/PcaGeneralDb1.txt
    # Licensed under the MIT License, Copyright (c) 2022 DFIR Artifact Museum
    db0_file = absolute_path("_data/plugins/os/windows/amcache/pca/PcaGeneralDb0.txt")
    db1_file = absolute_path("_data/plugins/os/windows/amcache/pca/PcaGeneralDb1.txt")
    fs_win.map_file("windows/appcompat/pca/PcaGeneralDb0.txt", db0_file)
    fs_win.map_file("windows/appcompat/pca/PcaGeneralDb1.txt", db1_file)

    # To test path resolving
    fs_win.map_file_fh("C:\\Program Files\\freefilesync\\bin\\freefilesync_x64.exe", BytesIO())

    target_win.add_plugin(AmcachePlugin)

    with patch(
        "dissect.target.plugins.os.windows.env.EnvironmentVariablePlugin._get_system_env_vars",
        return_value={"%programfiles%": "C:\\Program Files", "%windir%": "c:\\Windows"},
    ):
        records = list(target_win.amcache.general())

        assert len(records) == 176
        assert records[0].ts == datetime.datetime(2022, 5, 12, 19, 48, 9, 548000, tzinfo=datetime.timezone.utc)
        assert records[0].path == "C:\\Program Files\\freefilesync\\freefilesync.exe"
        assert records[0].type == "ABNORMAL_PROCESS_EXIT"
        assert records[0].name == "freefilesync"
        assert records[0].copyright == "freefilesync.org"
        assert records[0].version == "11.20"
        assert records[0].program_id == "000617915288ba535b4198ae58be4d9e2a4200000904"
        assert records[0].exit_message == "Abnormal process exit with code 0x2"


def mock_read_key_subkeys(self: AmcachePlugin, key: str) -> Iterator[Mock]:
    base_values = {
        "AppxPackageFullName": "Microsoft.Microsoft3DViewer_7.2105.4012.0_x64__8wekyb3d8bbwe",
        "AppxPackageRelativeId": "Microsoft.Microsoft3DViewer",
        "BinFileVersion": "7.2105.4012.0",
        "BinProductVersion": "7.2105.4012.0",
        "BinaryType": "pe64_amd64",
        "Language": 0,
        "LinkDate": "05/04/2021 17:43:39",
        "LongPathHash": "3dviewer.exe|40f275349895ac70",
        "LowerCaseLongPath": "c:\\program files\\windowsapps\\microsoft.0_x64__8wekyb3d8bbwe\\3dviewer.exe",
        "Name": "3DViewer.exe",
        "OriginalFileName": "3dviewer.exe",
        "ProductName": "view 3d",
        "ProductVersion": "7.2105.4012.0",
        "ProgramId": "0000df892556c2f7a6b7fa69f7009b5c08cb00000904",
        "Publisher": "microsoft corporation",
        "Size": 19456,
        "Usn": 32259776,
        "Version": "7.2105.4012.0",
    }

    mock_values = []
    for key, value in base_values.items():
        mock_value = Mock()
        mock_value.name = key
        mock_value.value = value
        mock_values.append(mock_value)

    mock_value = Mock()
    mock_value.name = "FileId"
    mock_value.value = self._mock_file_id
    mock_values.append(mock_value)

    mock_entry = Mock()
    mock_entry.timestamp = datetime.datetime(2021, 12, 31, tzinfo=datetime.timezone.utc)
    mock_entry.values = Mock(return_value=mock_values)

    yield mock_entry


@pytest.mark.parametrize(
    ("test_file_id", "expected_file_id"),
    [
        ("00008e01cdeb9a1c23cee421a647f29c45f67623be97", "8e01cdeb9a1c23cee421a647f29c45f67623be97"),
        ("", None),
        (None, None),
    ],
)
@patch.object(AmcachePlugin, "read_key_subkeys", mock_read_key_subkeys)
def test_parse_inventory_application_file(
    target_win: Target, test_file_id: str | None, expected_file_id: str | None
) -> None:
    with patch("dissect.target.plugins.os.windows.amcache.ApplicationFileAppcompatRecord") as mock_record:
        amcache_plugin = AmcachePlugin(target_win)
        amcache_plugin._mock_file_id = test_file_id
        records = list(amcache_plugin.parse_inventory_application_file())

        assert len(records) == 1
        call_kwargs = mock_record.call_args.kwargs

        assert call_kwargs.get("digest", None) == (None, expected_file_id, None)


def test_amcache_install_entry(target_win: Target) -> None:
    amcache_install_plugin = AmcacheInstallPlugin(target_win)

    amcache_install_plugin.logs = Path(absolute_path("_data/plugins/os/windows/amcache/install"))

    entries = list(amcache_install_plugin.amcache_install())

    created_order = [
        r"SOFTWARE\Microsoft\Windows\CurrentVersion\Uninstall\7-Zip",
        r"C:\Program Files\7-Zip\7-zip.dll",
        r"C:\Program Files\7-Zip\7-zip32.dll",
        r"C:\Program Files\7-Zip\7z.dll",
        r"C:\Program Files\7-Zip\7z.exe",
        r"C:\Program Files\7-Zip\7zFM.exe",
        r"C:\Program Files\7-Zip\7zG.exe",
        r"C:\Program Files\7-Zip\Uninstall.exe",
    ]
    assert len(entries) == 8

    for create, entry in zip(created_order, entries, strict=False):
        assert str(entry.create) == create
        assert str(entry.path) == r"C:\Users\JohnCena"
        assert str(entry.longname) == r"7z2201-x64.exe"
        assert entry.filesize == 1575742
        assert entry.size == 0x180B3E
        assert entry.pe_checksum == 0



from __future__ import annotations

from typing import TYPE_CHECKING, Any, Union

from dissect.database.ese.tools import certlog
from dissect.database.exception import Error

from dissect.target.exceptions import UnsupportedPluginError
from dissect.target.helpers.record import TargetRecordDescriptor
from dissect.target.plugin import Plugin, export

if TYPE_CHECKING:
    from collections.abc import Callable, Iterator
    from pathlib import Path

    from dissect.target.target import Target

RequestAttributeRecord = TargetRecordDescriptor(
    "filesystem/windows/certlog/request_attribute",
    [
        ("string", "attribute_name"),
        ("string", "common_name"),
        ("varint", "request_id"),
        ("string", "table_name"),
        ("string", "ca"),
        ("path", "source"),
    ],
)

CertificateExtensionRecord = TargetRecordDescriptor(
    "filesystem/windows/certlog/certificate_extension",
    [
        ("varint", "extension_flags"),
        ("string", "extension_name"),
        ("bytes", "extension_raw_value"),
        ("varint", "request_id"),
        ("string", "table_name"),
        ("string", "ca"),
        ("path", "source"),
    ],
)

CertificateRecord = TargetRecordDescriptor(
    "filesystem/windows/certlog/certificate",
    [
        ("digest", "fingerprint"),
        ("string", "certificate_template"),
        ("string", "common_name"),
        ("string", "country"),
        ("string", "device_serial_number"),
        ("string", "subject_dn"),
        ("string", "domain_component"),
        ("string", "email"),
        ("string", "given_name"),
        ("string", "initials"),
        ("string", "locality"),
        ("string", "organization"),
        ("string", "organizational_unit"),
        ("string", "public_key_algorithm"),
        ("string", "serial_number_hex"),
        ("varint", "serial_number"),
        ("string", "state_or_province"),
        ("string", "street_address"),
        ("string", "subject_key_identifier"),
        ("string", "sur_name"),
        ("string", "title"),
        ("string", "unstructured_address"),
        ("string", "unstructured_name"),
        ("string", "upn"),
        ("string", "enrollment_flags"),
        ("string", "general_flags"),
        ("string", "issuer_name_id"),
        ("datetime", "not_valid_after"),
        ("datetime", "not_valid_before"),
        ("string", "private_key_flags"),
        ("string", "public_key"),
        ("varint", "public_key_length"),
        ("string", "public_key_params"),
        ("string", "publish_expired_cert_in_crl"),
        ("bytes", "raw_certificate"),
        ("bytes", "raw_name"),
        ("varint", "request_id"),
        ("string", "table_name"),
        ("string", "ca"),
        ("path", "source"),
    ],
)

RequestRecord = TargetRecordDescriptor(
    "filesystem/windows/certlog/request",
    [
        ("string", "attestation_challenge"),
        ("string", "caller_name"),
        ("string", "common_name"),
        ("string", "country"),
        ("string", "device_serial_number"),
        ("string", "disposition"),
        ("string", "disposition_message"),
        ("string", "subject_dn"),
        ("string", "domain_component"),
        ("string", "email"),
        ("string", "endorsement_certificate_hash"),
        ("string", "endorsement_key_hash"),
        ("string", "given_name"),
        ("string", "initials"),
        ("string", "key_recovery_hashes"),
        ("string", "locality"),
        ("string", "organization"),
        ("string", "organizational_unit"),
        ("string", "raw_archived_key"),
        ("bytes", "raw_name"),
        ("bytes", "raw_old_certificate"),
        ("bytes", "raw_precertificate"),
        ("bytes", "raw_request"),
        ("string", "request_attributes"),
        ("string", "requester_name"),
        ("string", "request_flags"),
        ("varint", "request_id"),
        ("string", "request_type"),
        ("datetime", "resolved_when"),
        ("string", "revoked_effective_when"),
        ("string", "revoked_reason"),
        ("datetime", "revoked_when"),
        ("string", "signer_application_policies"),
        ("string", "signer_policies"),
        ("string", "state_or_province"),
        ("string", "request_status_code"),
        ("string", "street_address"),
        ("datetime", "submitted_when"),
        ("string", "sur_name"),
        ("string", "title"),
        ("string", "unstructured_address"),
        ("string", "unstructured_name"),
        ("string", "table_name"),
        ("string", "ca"),
        ("path", "source"),
    ],
)

CRLRecord = TargetRecordDescriptor(
    "filesystem/windows/certlog/crl",
    [
        ("varint", "count"),
        ("datetime", "crl_last_published"),
        ("varint", "crl_publish_attempts"),
        ("string", "crl_publish_error"),
        ("varint", "crl_publish_flags"),
        ("varint", "crl_publish_status_code"),
        ("datetime", "ts_effective"),
        ("string", "min_base"),
        ("varint", "name_id"),
        ("datetime", "next_publish"),
        ("datetime", "next_update"),
        ("varint", "number"),
        ("datetime", "propagation_complete"),
        ("bytes", "raw_crl"),
        ("varint", "row_id"),
        ("datetime", "this_publish"),
        ("datetime", "this_update"),
        ("string", "table_name"),
        ("string", "ca"),
        ("path", "source"),
    ],
)

CertLogRecord = Union[  # noqa: UP007
    RequestRecord, RequestAttributeRecord, CertificateRecord, CRLRecord, CertificateExtensionRecord
]

# {i: "".join("_" + c.lower() if c.isupper() else c for c in i.replace('$', ''))[1:] for i in a}
FIELD_MAPPINGS = {
    "$AttributeName": "attribute_name",
    "$AttributeValue": "common_name",
    "$CRLPublishError": "crl_publish_error",
    "$CallerName": "caller_name",
    "$CertificateHash": "fingerprint",
    "$CertificateHash2": "fingerprint",
    "$CertificateTemplate": "certificate_template",
    "$CommonName": "common_name",
    "$Country": "country",
    "$DeviceSerialNumber": "device_serial_number",
    "$DispositionMessage": "disposition_message",
    "$DistinguishedName": "subject_dn",
    "$DomainComponent": "domain_component",
    "$EMail": "email",
    "$EndorsementCertificateHash": "endorsement_certificate_hash",
    "$EndorsementKeyHash": "endorsement_key_hash",
    "$ExtensionName": "extension_name",
    "$GivenName": "given_name",
    "$Initials": "initials",
    "$KeyRecoveryHashes": "key_recovery_hashes",
    "$Locality": "locality",
    "$Organization": "organization",
    "$OrganizationalUnit": "organizational_unit",
    "$PublicKeyAlgorithm": "public_key_algorithm",
    "$RequestAttributes": "request_attributes",
    "$RequesterName": "requester_name",
    "$SerialNumber": "serial_number_hex",
    "$SignerApplicationPolicies": "signer_application_policies",
    "$SignerPolicies": "signer_policies",
    "$StateOrProvince": "state_or_province",
    "$StreetAddress": "street_address",
    "$SubjectKeyIdentifier": "subject_key_identifier",
    "$SurName": "sur_name",
    "$Title": "title",
    "$UPN": "upn",
    "$UnstructuredAddress": "unstructured_address",
    "$UnstructuredName": "unstructured_name",
    "AttestationChallenge": "attestation_challenge",
    "CRLLastPublished": "crl_last_published",
    "CRLPublishAttempts": "crl_publish_attempts",
    "CRLPublishFlags": "crl_publish_flags",
    "CRLPublishStatusCode": "crl_publish_status_code",
    "Count": "count",
    "Disposition": "disposition",
    "Effective": "ts_effective",  # prevent record_field_inconsistency
    "EnrollmentFlags": "enrollment_flags",
    "ExtensionFlags": "extension_flags",
    "ExtensionRawValue": "extension_raw_value",
    "GeneralFlags": "general_flags",
    "IssuerNameID": "issuer_name_id",
    "MinBase": "min_base",
    "NameId": "name_id",
    "NextPublish": "next_publish",
    "NextUpdate": "next_update",
    "NotAfter": "not_valid_after",
    "NotBefore": "not_valid_before",
    "Number": "number",
    "PrivateKeyFlags": "private_key_flags",
    "PropagationComplete": "propagation_complete",
    "PropgationComplete": "propagation_complete",  # typo in some Windows versions
    "PublicKey": "public_key",
    "PublicKeyLength": "public_key_length",
    "PublicKeyParams": "public_key_params",
    "PublishExpiredCertInCRL": "publish_expired_cert_in_crl",
    "RawArchivedKey": "raw_archived_key",
    "RawCRL": "raw_crl",
    "RawCertificate": "raw_certificate",
    "RawName": "raw_name",
    "RawOldCertificate": "raw_old_certificate",
    "RawPrecertificate": "raw_precertificate",
    "RawRequest": "raw_request",
    "RequestFlags": "request_flags",
    "RequestID": "request_id",
    "RequestType": "request_type",
    "ResolvedWhen": "resolved_when",
    "RevokedEffectiveWhen": "revoked_effective_when",
    "RevokedReason": "revoked_reason",
    "RevokedWhen": "revoked_when",
    "RowId": "row_id",
    "StatusCode": "request_status_code",  # prevent record_field_inconsistency
    "SubmittedWhen": "submitted_when",
    "TableName": "table_name",
    "ThisPublish": "this_publish",
    "ThisUpdate": "this_update",
}


def format_fingerprint(input_hash: str | None) -> tuple[str | None, str | None, str | None]:
    if input_hash:
        input_hash = input_hash.replace(" ", "")
        # hash is expected to be a sha1, but as it not documented, we make this function more flexible if hash is
        # in another standard format (md5/sha256), especially in the future
        match len(input_hash):
            case 32:
                return input_hash, None, None
            case 40:
                return None, input_hash, None
            case 64:
                return None, None, input_hash
            case _:
                raise ValueError(
                    "Unexpected hash size found while processing certlog "
                    f"$CertificateHash/$CertificateHash2 column: len {len(input_hash)}, content {input_hash}"
                )
    return None, None, None


def format_serial_number(serial_number_as_hex: str | None) -> str | None:
    if not serial_number_as_hex:
        return None
    return serial_number_as_hex.replace(" ", "")


def serial_number_as_int(serial_number_as_hex: str | None) -> int | None:
    if not serial_number_as_hex:
        return None
    return int(serial_number_as_hex, 16)


FORMATING_FUNC: dict[str, Callable[[Any], Any]] = {
    "fingerprint": format_fingerprint,
    "serial_number_hex": format_serial_number,
}


class CertLogPlugin(Plugin):
    """Return all available data stored in CertLog databases.

    Certificate Authority databases are databases related to the Active Directory Certificate Services (AD CS) feature.

    References:
        - https://learn.microsoft.com/en-us/openspecs/windows_protocols/ms-csra/5f06c74c-1a29-4fdf-b8dd-ae3300d1b90d
        - https://assets.crowdstrike.com/is/content/crowdstrikeinc/investigating-active-directory-certificate-abusepdf
        - https://learn.microsoft.com/en-gb/troubleshoot/windows-server/certificates-and-public-key-infrastructure-pki/move-certificate-server-database-log-files
    """

    __namespace__ = "certlog"

    def __init__(self, target: Target):
        super().__init__(target)
        self._certlog_dbs: list[tuple[certlog.CertLog, Path]] = []
        for path in self.get_paths():
            try:
                self._certlog_dbs.append((certlog.CertLog(path.open(mode="rb")), path))
            except Error as e:  # noqa: PERF203
                self.target.log.warning("Error opening Certlog database")
                self.target.log.debug("", exc_info=e)

    def _get_paths(self) -> Iterator[Path]:
        """Return all artifact files of interest to the plugin."""
        certlog_dir = self.target.resolve("%windir%/System32/CertLog")
        if certlog_dir.exists():
            yield from certlog_dir.glob("*.edb")

    def check_compatible(self) -> None:
        if not self._certlog_dbs:
            raise UnsupportedPluginError("No Certificate Authority Databases found")

    def read_records(self, table_name: str, record_type: CertLogRecord) -> Iterator[CertLogRecord]:
        for db, path in self._certlog_dbs:
            ca_name = path.stem
            table = next((table for table in db.tables() if table.name == table_name), None)
            if not table:
                self.target.log.warning("Table not found for ca %s: %s", ca_name, table_name)
                continue
            columns = [c.name for c in table.columns]
            for entry in db.records(table_name=table_name):
                values = (entry[name] for name in columns)
                column_values = zip(columns, values, strict=False)

                record_values = {}
                for column, value in column_values:
                    new_column = FIELD_MAPPINGS.get(column)
                    if new_column in FORMATING_FUNC:
                        try:
                            value = FORMATING_FUNC[new_column](value)
                        except Exception as e:
                            self.target.log.warning("Error formatting column %s (%s): %s", new_column, column, value)
                            self.target.log.debug("", exc_info=e)
                            value = None
                    if new_column and new_column not in record_values:
                        record_values[new_column] = value
                        # Serial number is format as int and string, to ease search of a specific sn in both format
                        if new_column == "serial_number_hex":
                            record_values["serial_number"] = serial_number_as_int(value)
                    elif new_column:
                        self.target.log.debug(
                            "Unexpected element while processing %s entries : %s column already exists "
                            "(mapped from original column name %s). This may be cause by two column that were not"
                            " expected to be present in the same time.",
                            table_name,
                            new_column,
                            column,
                        )
                    else:
                        self.target.log.debug(
                            "Unexpected column for table %s in CA %s: %s", table_name, ca_name, column
                        )

                yield record_type(
                    _target=self.target,
                    source=path,
                    table_name=table_name,
                    ca=ca_name,
                    **record_values,
                )

    @export(record=RequestRecord)
    def requests(self) -> Iterator[RequestRecord]:
        """Return the contents of the ``Requests`` table from all Certificate Authority databases.

        Gives insight into certificates requested (caller name, request ID, request attributes).
        """
        yield from self.read_records("Requests", RequestRecord)

    @export(record=RequestAttributeRecord)
    def request_attributes(self) -> Iterator[RequestAttributeRecord]:
        """Return the contents of the ``RequestAttributes`` table from all Certificate Authority databases.

        Gives insight into attributes of requested certificates (same information as in ``request_attributes`` field
        of ``Requests`` table).
        """
        yield from self.read_records("RequestAttributes", RequestAttributeRecord)

    @export(record=CRLRecord)
    def crls(self) -> Iterator[CRLRecord]:
        """Return the contents of the ``CRLs`` table from all Certificate Authority databases.

        Gives insight into the Certificate Revocation List of a Certificate Authority.
        """
        yield from self.read_records("CRLs", CRLRecord)

    @export(record=CertificateRecord)
    def certificates(self) -> Iterator[CertificateRecord]:
        """Return the contents of ``Certificates`` table from all Certificate Authority databases.

        Gives insight into issued certificates for a Certificate authority (public key, validity date).
        """
        yield from self.read_records("Certificates", CertificateRecord)

    @export(record=CertificateExtensionRecord)
    def certificate_extensions(self) -> Iterator[CertificateExtensionRecord]:
        """Return the contents of ``CertificateExtensions`` table from all Certificate Authority databases.

        Gives insight into certificate extensions for a CA.
        """
        yield from self.read_records("CertificateExtensions", CertificateExtensionRecord)



from flow.record.fieldtypes import datetime as dt

from dissect.target.filesystem import VirtualFilesystem
from dissect.target.plugins.os.unix.linux.recentlyused import RecentlyUsedPlugin
from dissect.target.target import Target
from tests._utils import absolute_path


def test_recently_used(target_unix_users: Target, fs_unix: VirtualFilesystem) -> None:
    """Verify parsing of valid recently-used.xbel file"""
    data_file = absolute_path("_data/plugins/os/unix/linux/recently-used.xbel")
    fs_unix.map_file("/home/user/.local/share/recently-used.xbel", data_file)
    target_unix_users.add_plugin(RecentlyUsedPlugin)

    # Check expected length
    results = list(target_unix_users.recently_used())
    assert len(results) == 15

    # Sanity checks on first parsed record
    assert results[0].user == "user"
    assert results[0].href == "file:///home/sjaak/.profile"
    assert results[0].ts == dt("2023-10-18 13:12:41.905277Z")
    assert results[0].added == dt("2023-10-18 13:12:41.905276Z")
    assert results[0].modified == dt("2023-10-18 13:14:09.483576Z")
    assert results[0].visited == dt("2023-10-18 13:12:41.905277Z")
    assert results[0].mimetype == "text/plain"


def test_recently_used_invalid_data(target_unix_users: Target, fs_unix: VirtualFilesystem) -> None:
    """Verify that the plugin doesn't error out on invalid data"""
    data_file = absolute_path("_data/plugins/os/unix/linux/recently-used-invalid-data.xbel")
    fs_unix.map_file("/home/user/.local/share/recently-used.xbel", data_file)
    target_unix_users.add_plugin(RecentlyUsedPlugin)

    results = list(target_unix_users.recently_used())
    assert len(results) == 15


def test_recently_used_broken_xml(target_unix_users: Target, fs_unix: VirtualFilesystem) -> None:
    """Verify that the plugin doesn't error out on broken XML structure"""
    data_file = absolute_path("_data/plugins/os/unix/linux/recently-used-broken-xml.xbel")
    fs_unix.map_file("/home/user/.local/share/recently-used.xbel", data_file)
    target_unix_users.add_plugin(RecentlyUsedPlugin)

    results = list(target_unix_users.recently_used())
    assert len(results) == 0



from __future__ import annotations

import lzma
import sys
from typing import TYPE_CHECKING, Any, BinaryIO

from dissect.cstruct import cstruct
from dissect.util import ts
from dissect.util.compression import lz4

from dissect.target.exceptions import UnsupportedPluginError
from dissect.target.helpers.logging import get_logger
from dissect.target.helpers.record import TargetRecordDescriptor
from dissect.target.helpers.utils import IntEnumMissing
from dissect.target.plugin import Plugin, export

if TYPE_CHECKING:
    from collections.abc import Callable, Iterator

    from dissect.target.target import Target

if sys.version_info >= (3, 14):
    from compression import zstd  # novermin
else:
    from backports import zstd


log = get_logger(__name__)

# The events have undocumented fields that are not part of the record
JournalRecord = TargetRecordDescriptor(
    "linux/log/journal",
    [
        ("datetime", "ts"),
        # User Journal fields
        ("string", "message"),
        ("string", "message_id"),
        ("string", "priority"),
        ("path", "code_file"),
        ("varint", "code_line"),
        ("string", "code_func"),
        ("varint", "errno"),
        ("string", "invocation_id"),
        ("string", "user_invocation_id"),
        ("string", "syslog_facility"),
        ("string", "syslog_identifier"),
        ("varint", "syslog_pid"),
        ("string", "syslog_raw"),
        ("string", "documentation"),
        ("varint", "tid"),
        ("string", "unit"),
        ("string", "user_unit"),
        # Trusted Journal fields
        ("varint", "pid"),
        ("varint", "uid"),
        ("varint", "gid"),
        ("string", "comm"),
        ("path", "exe"),
        ("string", "cmdline"),
        ("string", "cap_effective"),
        ("varint", "audit_session"),
        ("varint", "audit_loginuid"),
        ("path", "systemd_cgroup"),
        ("string", "systemd_slice"),
        ("string", "systemd_unit"),
        ("string", "systemd_user_unit"),
        ("string", "systemd_user_slice"),
        ("string", "systemd_session"),
        ("string", "systemd_owner_uid"),
        ("string", "selinux_context"),
        ("string", "kernel_boot_id"),
        ("string", "machine_id"),
        ("string", "systemd_invocation_id"),
        ("string", "transport"),
        ("string", "stream_id"),
        ("string", "line_break"),
        ("string", "namespace"),
        ("string", "runtime_scope"),
        # Kernel Journal fields
        ("string", "kernel_device"),
        ("string", "kernel_subsystem"),
        ("string", "udev_sysname"),
        ("path", "udev_devnode"),
        ("path", "udev_devlink"),
        # Other fields
        ("string", "journal_hostname"),
        ("path", "source"),
    ],
)

journal_def = """
#define HEADER_SIGNATURE b"LPKSHHRH"

typedef uint8 uint8_t;
typedef uint32 le32_t;
typedef uint64 le64_t;

enum State : uint8_t {
    OFFLINE   =  0,
    ONLINE    =  1,
    ARCHIVED  =  2,
    UNKNOWN
};

union sd_id128_t {
    uint8_t   bytes[16];
    uint64_t  qwords[2];
};

flag IncompatibleFlag : le32_t {
    HEADER_INCOMPATIBLE_COMPRESSED_XZ   = 1,
    HEADER_INCOMPATIBLE_COMPRESSED_LZ4  = 2,
    HEADER_INCOMPATIBLE_KEYED_HASH      = 4,
    HEADER_INCOMPATIBLE_COMPRESSED_ZSTD = 8,
    HEADER_INCOMPATIBLE_COMPACT         = 16,             // indicates that the Journal file uses the new binary format
};

struct Header {
    char              signature[8];
    le32_t            compatible_flags;
    IncompatibleFlag  incompatible_flags;
    State             state;
    uint8_t           reserved[7];
    sd_id128_t        file_id;
    sd_id128_t        machine_id;
    sd_id128_t        tail_entry_boot_id;
    sd_id128_t        seqnum_id;
    le64_t            header_size;
    le64_t            arena_size;
    le64_t            data_hash_table_offset;
    le64_t            data_hash_table_size;
    le64_t            field_hash_table_offset;
    le64_t            field_hash_table_size;
    le64_t            tail_object_offset;
    le64_t            n_objects;
    le64_t            n_entries;
    le64_t            tail_entry_seqnum;
    le64_t            head_entry_seqnum;
    le64_t            entry_array_offset;
    le64_t            head_entry_realtime;
    le64_t            tail_entry_realtime;
    le64_t            tail_entry_monotonic;
    le64_t            n_data;
    le64_t            n_fields;
    le64_t            n_tags;
    le64_t            n_entry_arrays;
    le64_t            data_hash_chain_depth;
    le64_t            field_hash_chain_depth;
    le32_t            tail_entry_array_offset;
    le32_t            tail_entry_array_n_entries;
    le64_t            tail_entry_offset;
};

enum ObjectType : uint8 {
    OBJECT_UNUSED,
    OBJECT_DATA,
    OBJECT_FIELD,
    OBJECT_ENTRY,
    OBJECT_DATA_HASH_TABLE,
    OBJECT_FIELD_HASH_TABLE,
    OBJECT_ENTRY_ARRAY,
    OBJECT_TAG,
    _OBJECT_TYPE_MAX
};

flag ObjectFlag : uint8 {
    OBJECT_UNCOMPRESSED      =  0,
    OBJECT_COMPRESSED_XZ     =  1,
    OBJECT_COMPRESSED_LZ4    =  2,
    OBJECT_COMPRESSED_ZSTD   =  4,
    _OBJECT_COMPRESSED_MASK  =  7
};

struct ObjectHeader {
    ObjectType  type;                                     // The type field is one of the object types listed above
    uint8_t     flags;                                    // If DATA object the value is ObjectFlag
    uint8_t     reserved[6];
    le64_t      size;                                     // The size field encodes the size of the object including all its headers and payload
};


// The first four members are copied from ObjectHeader, so that the size can be used as the length of payload
struct DataObject {
    // ObjectType  type;
    ObjectFlag  flags;
    uint8_t     reserved[6];
    le64_t      size;
    le64_t      hash;
    le64_t      next_hash_offset;
    le64_t      next_field_offset;
    le64_t      entry_offset;
    le64_t      entry_array_offset;
    le64_t      n_entries;
    char        payload[size - 64];                       // Data objects carry actual field data in the payload[] array.
};

// If the HEADER_INCOMPATIBLE_COMPACT flag is set, two extra fields are stored to allow immediate access
// to the tail entry array in the DATA object's entry array chain.
struct DataObject_Compact {
    // ObjectType  type;
    ObjectFlag  flags;
    uint8_t     reserved[6];
    le64_t      size;
    le64_t      hash;
    le64_t      next_hash_offset;
    le64_t      next_field_offset;
    le64_t      entry_offset;
    le64_t      entry_array_offset;
    le64_t      n_entries;
    le32_t      tail_entry_array_offset;
    le32_t      tail_entry_array_n_entries;
    char        payload[size - 72];                       // Data objects carry actual field data in the payload[] array.
};

struct EntryItem {
    le64_t object_offset;
    le64_t hash;
};

struct EntryItem_Compact {
    le32_t object_offset;
}

// The first four members are copied from ObjectHeader, so that the size can be used as the length of items
struct EntryObject {
    ObjectType  type;
    uint8_t     flags;
    uint8_t     reserved[6];
    le64_t      size;
    le64_t      seqnum;
    le64_t      realtime;
    le64_t      monotonic;
    sd_id128_t  boot_id;
    le64_t      xor_hash;
    EntryItem   items[(size - 64) / 16];                  // The size minus the previous members divided by the size of the items
};

// If the HEADER_INCOMPATIBLE_COMPACT flag is set, DATA object offsets are stored as 32-bit integers instead of 64bit
// and the unused hash field per data object is not stored anymore.
struct EntryObject_Compact {
    ObjectType  type;
    uint8_t     flags;
    uint8_t     reserved[6];
    le64_t      size;
    le64_t      seqnum;
    le64_t      realtime;
    le64_t      monotonic;
    sd_id128_t  boot_id;
    le64_t      xor_hash;
    EntryItem_Compact   items[(size - 64) / 4];
};

// The first four members are copied from from ObjectHeader, so that the size can be used as the length of entry_object_offsets
struct EntryArrayObject {
    // ObjectType  type;
    uint8_t     flags;
    uint8_t     reserved[6];
    le64_t      size;
    le64_t      next_entry_array_offset;
    le64_t      entry_object_offsets[(size - 24) / 8];    // The size minus the previous members divided by the size of the offset
};

struct EntryArrayObject_Compact {
    // ObjectType  type;
    uint8_t     flags;
    uint8_t     reserved[6];
    le64_t      size;
    le64_t      next_entry_array_offset;
    le32_t      entry_object_offsets[(size - 24) / 4];
};
"""  # noqa: E501

c_journal = cstruct().load(journal_def)


class JournalMessagePriority(IntEnumMissing):
    """Journal message priority enum.

    References:
        - https://wiki.archlinux.org/title/Systemd/Journal
    """

    EMERG = 0
    ALERT = 1
    CRIT = 2
    ERR = 3
    WARNING = 4
    NOTICE = 5
    INFO = 6
    DEBUG = 7


def get_optional(value: str, to_type: Callable) -> Any | None:
    """Return the value if True, otherwise return None."""

    if not value:
        return None

    try:
        return to_type(value)

    except ValueError as e:
        log.error("Unable to cast '%s' to %s", value, to_type)  # noqa: TRY400
        log.debug("", exc_info=e)
        return None


# Sometimes stringy None is inserted by external tools like Ansible
def int_or_none(value: str) -> int | None:
    return int(value) if value and value != "None" else None


class JournalFile:
    """Parse Systemd Journal file format.

    References:
        - https://github.com/systemd/systemd/blob/206f0f397edf1144c63a158fb30f496c3e89f256/docs/JOURNAL_FILE_FORMAT.md
        - https://github.com/libyal/dtformats/blob/c4fc2b8102702c64b58f145971821986bf74e6c0/documentation/Systemd%20journal%20file%20format.asciidoc
    """

    def __init__(self, fh: BinaryIO, target: Target):
        self.fh = fh
        self.target = target

        try:
            self.header = c_journal.Header(self.fh)
        except EOFError as e:
            raise ValueError(f"Invalid systemd Journal file: {e}")

        if self.header.signature != c_journal.HEADER_SIGNATURE:
            raise ValueError(f"Journal file has invalid magic header: {self.header.signature!r}'")

    def decode_value(self, value: bytes) -> tuple[str, str]:
        """Decode the given bytes to a key value pair."""
        value = value.decode(errors="surrogateescape").strip().lstrip("_")
        key, value = value.split("=", 1)
        key = key.lower()
        return key, value

    def __iter__(self) -> Iterator[dict[str, int | str]]:
        "Iterate over the entry objects to read payloads."

        offset = self.header.entry_array_offset
        while offset != 0:
            self.fh.seek(offset)
            object_type = self.fh.read(1)[0]

            if object_type == c_journal.ObjectType.OBJECT_UNUSED:
                self.target.log.warning(
                    "ObjectType OBJECT_UNUSED encountered for next OBJECT_ENTRY_ARRAY offset at 0x%X. "
                    "This indicates allocated space in the journal file which is not used yet.",
                    offset,
                )
                break

            elif object_type != c_journal.ObjectType.OBJECT_ENTRY_ARRAY:
                raise ValueError(f"Expected OBJECT_ENTRY_ARRAY or OBJECT_UNUSED at offset {offset}")

            if self.header.incompatible_flags & c_journal.IncompatibleFlag.HEADER_INCOMPATIBLE_COMPACT:
                entry_array_object = c_journal.EntryArrayObject_Compact(self.fh)
            else:
                entry_array_object = c_journal.EntryArrayObject(self.fh)

            for entry_object_offset in entry_array_object.entry_object_offsets:
                if entry_object_offset:
                    yield from self._parse_entry_object(offset=entry_object_offset)

            offset = entry_array_object.next_entry_array_offset

    def _parse_entry_object(self, offset: int) -> Iterator[dict]:
        self.fh.seek(offset)

        try:
            if self.header.incompatible_flags & c_journal.IncompatibleFlag.HEADER_INCOMPATIBLE_COMPACT:
                entry = c_journal.EntryObject_Compact(self.fh)
            else:
                entry = c_journal.EntryObject(self.fh)

        except EOFError as e:
            self.target.log.warning("Unable to read Journal EntryObject at offset %s in: %s", offset, self.fh)
            self.target.log.debug("", exc_info=e)
            return

        event = {"ts": ts.from_unix_us(entry.realtime)}
        for item in entry.items:
            try:
                self.fh.seek(item.object_offset)

                if self.fh.read(1)[0] != c_journal.ObjectType.OBJECT_DATA:
                    continue

                if self.header.incompatible_flags & c_journal.IncompatibleFlag.HEADER_INCOMPATIBLE_COMPACT:
                    data_object = c_journal.DataObject_Compact(self.fh)
                else:
                    data_object = c_journal.DataObject(self.fh)

                if not data_object.payload:
                    continue

                data = data_object.payload

                if data_object.flags & c_journal.ObjectFlag.OBJECT_COMPRESSED_XZ:
                    data = lzma.decompress(data)

                elif data_object.flags & c_journal.ObjectFlag.OBJECT_COMPRESSED_LZ4:
                    data = lz4.decompress(data[8:])

                elif data_object.flags & c_journal.ObjectFlag.OBJECT_COMPRESSED_ZSTD:
                    data = zstd.decompress(data)

                key, value = self.decode_value(data)
                event[key] = value

            except Exception as e:
                self.target.log.warning(
                    "Journal DataObject could not be parsed at offset %s in %s",
                    item.object_offset,
                    getattr(self.fh, "name", None),
                )
                self.target.log.debug("", exc_info=e)
                continue

        yield event


class JournalPlugin(Plugin):
    """Systemd Journal plugin."""

    JOURNAL_PATHS = ("/var/log/journal",)  # TODO: /run/systemd/journal
    JOURNAL_GLOB = "*/*.journal*"  # The extensions .journal and .journal~

    def __init__(self, target: Target):
        super().__init__(target)
        self.journal_files = []

        for journal_path in self.JOURNAL_PATHS:
            self.journal_files.extend(self.target.fs.path(journal_path).glob(self.JOURNAL_GLOB))

    def check_compatible(self) -> None:
        if not self.journal_files:
            raise UnsupportedPluginError("No journald files found")

    @export(record=JournalRecord)
    def journal(self) -> Iterator[JournalRecord]:
        """Return the contents of Systemd Journal log files.

        References:
            - https://wiki.archlinux.org/title/Systemd/Journal
            - https://github.com/systemd/systemd/blob/9203abf79f1d05fdef9b039e7addf9fc5a27752d/man/systemd.journal-fields.xml
        """
        path_function = self.target.fs.path

        for journal_file in self.journal_files:
            if not journal_file.is_file():
                self.target.log.warning("Unable to parse journal file as it is not a file: %s", journal_file)
                continue

            try:
                fh = journal_file.open()
                journal = JournalFile(fh, self.target)

            except Exception as e:
                self.target.log.warning("Unable to parse journal file structure: %s: %s", journal_file, str(e))
                self.target.log.debug("", exc_info=e)
                continue

            for entry in journal:
                priority = int_or_none(entry.get("priority"))
                yield JournalRecord(
                    ts=entry.get("ts"),
                    message=entry.get("message"),
                    message_id=entry.get("message_id"),
                    priority=JournalMessagePriority(priority).name.lower() if priority else None,
                    code_file=get_optional(entry.get("code_file"), path_function),
                    code_line=int_or_none(entry.get("code_line")),
                    code_func=entry.get("code_func"),
                    errno=int_or_none(entry.get("errno")),
                    invocation_id=entry.get("invocation_id"),
                    user_invocation_id=entry.get("user_invocation_id"),
                    syslog_facility=entry.get("syslog_facility"),
                    syslog_identifier=entry.get("syslog_identifier"),
                    syslog_pid=int_or_none(entry.get("syslog_pid")),
                    syslog_raw=entry.get("syslog_raw"),
                    documentation=entry.get("documentation"),
                    tid=int_or_none(entry.get("tid")),
                    unit=entry.get("unit"),
                    user_unit=entry.get("user_unit"),
                    pid=int_or_none(entry.get("pid")),
                    uid=int_or_none(entry.get("uid")),
                    gid=int_or_none(entry.get("gid")),
                    comm=entry.get("comm"),
                    exe=get_optional(entry.get("exe"), path_function),
                    cmdline=entry.get("cmdline"),
                    cap_effective=entry.get("cap_effective"),
                    audit_session=int_or_none(entry.get("audit_session")),
                    audit_loginuid=int_or_none(entry.get("audit_loginuid")),
                    systemd_cgroup=get_optional(entry.get("systemd_cgroup"), path_function),
                    systemd_slice=entry.get("systemd_slice"),
                    systemd_unit=entry.get("systemd_unit"),
                    systemd_user_unit=entry.get("systemd_user_unit"),
                    systemd_user_slice=entry.get("systemd_user_slice"),
                    systemd_session=entry.get("systemd_session"),
                    systemd_owner_uid=entry.get("systemd_owner_uid"),
                    selinux_context=entry.get("selinux_context"),
                    kernel_boot_id=entry.get("boot_id"),
                    machine_id=entry.get("machine_id"),
                    systemd_invocation_id=entry.get("systemd_invocation_id"),
                    transport=entry.get("transport"),
                    stream_id=entry.get("stream_id"),
                    line_break=entry.get("line_break"),
                    namespace=entry.get("namespace"),
                    runtime_scope=entry.get("runtime_scope"),
                    kernel_device=entry.get("kernel_device"),
                    kernel_subsystem=entry.get("kernel_subsystem"),
                    udev_sysname=entry.get("udev_sysname"),
                    udev_devnode=get_optional(entry.get("udev_devnode"), path_function),
                    udev_devlink=get_optional(entry.get("udev_devlink"), path_function),
                    journal_hostname=entry.get("hostname"),
                    source=journal_file,
                    _target=self.target,
                )



from __future__ import annotations

import ast
import importlib.machinery
import importlib.util
from pathlib import Path
from typing import TYPE_CHECKING

from dissect.target.helpers.logging import get_logger

if TYPE_CHECKING:
    from types import ModuleType


log = get_logger(__name__)

CONFIG_NAME = ".targetcfg.py"


def load(paths: list[Path | str] | Path | str | None) -> ModuleType:
    """Attempt to load one configuration from the provided path(s)."""

    if isinstance(paths, (Path, str)):
        paths = [paths]

    config_spec = importlib.machinery.ModuleSpec("config", None)
    config = importlib.util.module_from_spec(config_spec)
    config_file = _find_config_file(paths)

    if config_file:
        config_values = _parse_ast(config_file.read_bytes())
        config.__dict__.update(config_values)

    return config


def _parse_ast(code: str) -> dict[str, str | int]:
    # Only allow basic value assignments for backwards compatibility
    obj = {}

    module = ast.parse(code)
    if not isinstance(module, ast.Module):
        log.debug("Config did not parse to a module AST -- skipping")
        return obj

    for statement in module.body:
        if (
            not isinstance(statement, ast.Assign)
            or len(statement.targets) != 1
            or not isinstance(statement.value, ast.Constant)
        ):
            log.debug("Skipping non-constant assignment")
            continue

        target = statement.targets[0]
        if not isinstance(target, ast.Name) or not isinstance(target.ctx, ast.Store):
            log.debug("Skipping non-name assignment store")
            continue

        obj[target.id] = statement.value.value

    return obj


def _find_config_file(paths: list[Path | str] | None) -> Path | None:
    """Find a config file anywhere in the given path(s) and return it.

    This algorithm allows parts of the path to not exist or the last part to be a filename.
    It also does not look in the root directory ('/') for config files.
    """

    if not paths:
        return None

    config_file = None

    for path in paths:
        if not path:
            continue

        path = Path(path)
        cur_path = path.absolute()

        # Look for a config file in provided path or in parent directories until found.
        # This will not allow config files in the root directory or "virtual" targets (e.g. "local").
        while not config_file and cur_path.exists() and cur_path.name != "":
            cur_config = cur_path.joinpath(CONFIG_NAME)
            if cur_config.is_file():
                config_file = cur_config
            cur_path = cur_path.parent

        if config_file:
            break

    return config_file



from __future__ import annotations

from typing import TYPE_CHECKING

from dissect.hypervisor import ovf

from dissect.target import container
from dissect.target.loader import Loader

if TYPE_CHECKING:
    from pathlib import Path

    from dissect.target.target import Target


class OvfLoader(Loader):
    """Load Open Virtualization Format (OVF) files.

    References:
        - https://en.wikipedia.org/wiki/Open_Virtualization_Format
    """

    def __init__(self, path: Path, **kwargs):
        super().__init__(path, **kwargs)

        with path.open("r") as fh:
            self.ovf = ovf.OVF(fh)

    @staticmethod
    def detect(path: Path) -> bool:
        return path.suffix.lower() == ".ovf"

    def map(self, target: Target) -> None:
        for disk in self.ovf.disks():
            disk = disk.replace("\\", "/")
            _, _, fname = disk.rpartition("/")
            path = self.base_path.joinpath(fname)

            try:
                target.disks.add(container.open(path))
            except Exception:
                target.log.exception("Failed to load disk: %s", disk)



from dissect.target.filesystem import VirtualFilesystem
from dissect.target.plugin import OperatingSystem
from dissect.target.plugins.os.unix.bsd.darwin.ios._os import IOSPlugin
from dissect.target.plugins.os.unix.bsd.darwin.ios.applications import (
    IOSApplicationsPlugin,
)
from dissect.target.plugins.os.unix.bsd.darwin.ios.locale import LocalePlugin
from dissect.target.target import Target
from tests._utils import absolute_path


def test_ios_detect(target_bare: Target, fs_ios: VirtualFilesystem) -> None:
    """Test if we correctly detect an iOS target."""
    target_bare.filesystems.add(fs_ios)
    target_bare.apply()
    assert IOSPlugin.detect(target_bare)
    assert target_bare._os_plugin is IOSPlugin


def test_ios_os(target_ios: Target, fs_ios: VirtualFilesystem) -> None:
    """Test if the iOS plugin works on an iOS target.

    Data based on example iOS image from Digital Corpora.

    References:
        - https://corp.digitalcorpora.org/corpora/mobile/iOS17/
    """

    fs_ios.map_dir("/", absolute_path("_data/plugins/os/unix/bsd/darwin/ios/_os"))
    target_ios.add_plugin(IOSPlugin)
    target_ios.add_plugin(LocalePlugin)
    target_ios.add_plugin(IOSApplicationsPlugin)
    target_ios.apply()

    assert target_ios.os == OperatingSystem.IOS
    assert target_ios.version == "iPhone OS 17.3 (21D50)"
    assert target_ios.architecture == "arm64-ios"
    assert target_ios.hostname == "This-Iss-iPhone"
    assert target_ios.timezone == "America/New_York"
    assert target_ios.language == ["en_US"]

    users = list(target_ios.users())
    assert len(users) == 43

    apps = list(target_ios.applications())
    assert len(apps) == 4

    assert sorted([app.name for app in apps]) == [
        # System apps
        "Books",
        "Calculator",
        "Compass",
        # AppStore apps
        "DuckDuckGo Private Browser",
    ]



from __future__ import annotations

from typing import TYPE_CHECKING, BinaryIO

from dissect.jffs import jffs2
from dissect.jffs.c_jffs2 import c_jffs2

from dissect.target.exceptions import (
    FileNotFoundError,
    FilesystemError,
    IsADirectoryError,
    NotADirectoryError,
    NotASymlinkError,
)
from dissect.target.filesystem import DirEntry, Filesystem, FilesystemEntry
from dissect.target.helpers import fsutil

if TYPE_CHECKING:
    from collections.abc import Iterator


class JffsFilesystem(Filesystem):
    __type__ = "jffs"

    def __init__(self, fh: BinaryIO, *args, **kwargs):
        super().__init__(fh, *args, **kwargs)
        self.jffs2 = jffs2.JFFS2(fh)

    @staticmethod
    def _detect(fh: BinaryIO) -> bool:
        return int.from_bytes(fh.read(2), "little") in (
            c_jffs2.JFFS2_MAGIC_BITMASK,
            c_jffs2.JFFS2_OLD_MAGIC_BITMASK,
        )

    def get(self, path: str) -> FilesystemEntry:
        return JffsFilesystemEntry(self, path, self._get_node(path))

    def _get_node(self, path: str, node: jffs2.INode | None = None) -> jffs2.INode:
        try:
            return self.jffs2.get(path, node)
        except jffs2.FileNotFoundError as e:
            raise FileNotFoundError(path) from e
        except jffs2.NotADirectoryError as e:
            raise NotADirectoryError(path) from e
        except jffs2.NotASymlinkError as e:
            raise NotASymlinkError(path) from e
        except jffs2.Error as e:
            raise FileNotFoundError(path) from e


class JffsDirEntry(DirEntry):
    fs: JffsFilesystem
    entry: jffs2.INode

    def get(self) -> JffsFilesystemEntry:
        return JffsFilesystemEntry(self.fs, self.path, self.entry)

    def stat(self, *, follow_symlinks: bool = True) -> fsutil.stat_result:
        return self.get().stat(follow_symlinks=follow_symlinks)


class JffsFilesystemEntry(FilesystemEntry):
    fs: JffsFilesystem
    entry: jffs2.INode

    def get(self, path: str) -> FilesystemEntry:
        entry_path = fsutil.join(self.path, path, alt_separator=self.fs.alt_separator)
        entry = self.fs._get_node(path, self.entry)
        return JffsFilesystemEntry(self.fs, entry_path, entry)

    def open(self) -> BinaryIO:
        if self.is_dir():
            raise IsADirectoryError(self.path)
        return self._resolve().entry.open()

    def scandir(self) -> Iterator[JffsDirEntry]:
        if not self.is_dir():
            raise NotADirectoryError(self.path)

        for name, entry in self._resolve().entry.iterdir():
            # TODO: Separate INode and DirEntry in dissect.jffs
            yield JffsDirEntry(self.fs, self.path, name, entry)

    def is_dir(self, follow_symlinks: bool = True) -> bool:
        try:
            return self._resolve(follow_symlinks).entry.is_dir()
        except FilesystemError:
            return False

    def is_file(self, follow_symlinks: bool = True) -> bool:
        try:
            return self._resolve(follow_symlinks).entry.is_file()
        except FilesystemError:
            return False

    def is_symlink(self) -> bool:
        return self.entry.is_symlink()

    def readlink(self) -> str:
        if not self.is_symlink():
            raise NotASymlinkError

        return self.entry.link

    def stat(self, follow_symlinks: bool = True) -> fsutil.stat_result:
        return self._resolve(follow_symlinks).lstat()

    def lstat(self) -> fsutil.stat_result:
        node = self.entry.inode

        # mode, ino, dev, nlink, uid, gid, size, atime, mtime, ctime
        st_info = fsutil.stat_result(
            [
                self.entry.mode,
                self.entry.inum,
                id(self.fs),
                self.entry.nlink,
                node.uid,
                node.gid,
                node.isize,
                self.entry.atime.timestamp(),
                self.entry.mtime.timestamp(),
                self.entry.ctime.timestamp(),
            ]
        )

        # JFFS2 block size is a function of the "erase size" of the underlying flash device.
        # Linux stat reports the default block size, which is defined as 4k in libc.
        st_info.st_blksize = 4096
        st_info.st_blocks = (node.isize + 511) // 512 if self.is_file() else 0

        return st_info



from __future__ import annotations

from datetime import datetime, timezone
from typing import TYPE_CHECKING

from dissect.target.plugins.os.unix.log.atop import AtopPlugin
from tests._utils import absolute_path

if TYPE_CHECKING:
    from dissect.target.filesystem import VirtualFilesystem
    from dissect.target.target import Target


def test_atop_plugin(target_unix: Target, fs_unix: VirtualFilesystem) -> None:
    data_file = absolute_path("_data/plugins/os/unix/log/atop/atop")
    fs_unix.map_file("var/log/atop/atop_20221111", data_file)

    target_unix.add_plugin(AtopPlugin)

    results = list(target_unix.atop())
    assert len(results) == 2219
    assert results[0].ts == datetime(2022, 11, 11, 19, 50, 44, tzinfo=timezone.utc)
    assert results[0].process == "systemd"
    assert results[0].cmdline == "/sbin/init"
    assert results[0].tgid == 1
    assert results[0].pid == 1
    assert results[0].ppid == 0
    assert results[0].ruid == 0
    assert results[0].euid == 0
    assert results[0].suid == 0
    assert results[0].fsuid == 0
    assert results[0].rgid == 0
    assert results[0].egid == 0
    assert results[0].sgid == 0
    assert results[0].fsgid == 0
    assert results[0].nthr == 1
    assert bool(results[0].isproc) is True
    assert results[0].state == "S"
    assert results[0].excode == -2147483648
    assert results[0].elaps == 0
    assert results[0].nthrslpi == 1
    assert results[0].nthrslpu == 0
    assert results[0].nthrrun == 0
    assert results[0].ctid == 0
    assert results[0].vpid == 0
    assert bool(results[0].wasinactive) is False
    assert results[0].container == ""
    assert str(results[0].filepath) == "/var/log/atop/atop_20221111"



from __future__ import annotations

from typing import TYPE_CHECKING

from defusedxml import ElementTree

from dissect.target.exceptions import UnsupportedPluginError
from dissect.target.helpers.record import TargetRecordDescriptor
from dissect.target.plugin import Plugin, export

if TYPE_CHECKING:
    from collections.abc import Iterator


VirtualMachineInventoryRecord = TargetRecordDescriptor(
    "esxi/vm/inventory",
    [
        ("path", "path"),
    ],
)

VirtualMachineOrphanRecord = TargetRecordDescriptor(
    "esxi/vm/orphaned",
    [
        ("path", "path"),
    ],
)


class VirtualMachinePlugin(Plugin):
    """Plugin to list ESXi virtual machines."""

    __namespace__ = "vm"

    def check_compatible(self) -> None:
        if self.target.os != "esxi":
            raise UnsupportedPluginError("ESXi specific plugin loaded on non-ESXi target")

        if not self.target.fs.path("/etc/vmware/hostd/vmInventory.xml").exists():
            raise UnsupportedPluginError("vmInventory.xml not found on target")

    @export(record=VirtualMachineInventoryRecord)
    def inventory(self) -> Iterator[VirtualMachineInventoryRecord]:
        """Yield all virtual machines registered on the ESXi host."""
        if (inv_file := self.target.fs.path("/etc/vmware/hostd/vmInventory.xml")).is_file():
            root = ElementTree.fromstring(inv_file.read_text())
            for entry in root.iter("ConfigEntry"):
                yield VirtualMachineInventoryRecord(
                    path=self.target.fs.path(entry.findtext("vmxCfgPath")),
                    _target=self.target,
                )

    @export(record=VirtualMachineOrphanRecord)
    def orphaned(self) -> Iterator[VirtualMachineOrphanRecord]:
        """Yield all virtual machines found at ``/vmfs/volumes/*/*/*.vmx`` that are NOT in the inventory.

        NOTE: If the target is part of a cluster, this may yield "false positives" for VMs registered
        on other hosts in the cluster.
        """
        inventory = {str(r.path) for r in self.inventory()}

        for vmx_path in self.target.fs.path("/vmfs/volumes").glob("*/*/*.vmx"):
            if str(vmx_path) not in inventory:
                yield VirtualMachineOrphanRecord(
                    path=vmx_path,
                    _target=self.target,
                )



from __future__ import annotations

from typing import TYPE_CHECKING

from dissect.target.exceptions import FileNotFoundError
from dissect.target.filesystems.dir import DirectoryFilesystem
from dissect.target.filesystems.ntfs import NtfsFilesystem
from dissect.target.loader import Loader

if TYPE_CHECKING:
    from pathlib import Path

    from dissect.target.target import Target


class VBLoader(Loader):
    @staticmethod
    def detect(path: Path) -> bool:
        mft_exists = path.joinpath("MFT_C.bin").exists()
        c_drive_exists = path.joinpath("C_drive").exists()
        config_exists = path.joinpath("Windows/System32/config").exists()

        return (mft_exists or c_drive_exists) and config_exists

    def map(self, target: Target) -> None:
        remap_overlay = target.fs.append_layer()
        ntfs_overlay = target.fs.append_layer()
        dfs = DirectoryFilesystem(self.path, case_sensitive=False)
        target.filesystems.add(dfs)

        for f in dfs.listdir_ext("/"):
            if not f.name.endswith("_drive"):
                continue
            fh_mft = f.get("MFT_C.bin").open()
            fh_boot = f.get("$Boot").open()
            fh_sds = f.get("$Secure[ADS_$SDS]").open()

            try:
                fh_usnjrnl = f.get("$UsnJrnl_$J.bin").open()
            except FileNotFoundError:
                fh_usnjrnl = None

            fs = NtfsFilesystem(mft=fh_mft, boot=fh_boot, sds=fh_sds, usnjrnl=fh_usnjrnl)
            target.filesystems.add(fs)

            if f.name == "C_drive":
                ntfs_overlay.mount("sysvol", fs)
                remap_overlay.map_file("sysvol/$mft", fh_mft, "sysvol/$mft")
                if fh_usnjrnl:
                    remap_overlay.map_file("sysvol/$extend/$usnjrnl:$J", fh_usnjrnl, "sysvol/$extend/$usnjrnl:$J")
            drive = f.name.split("_")[0].lower()
            drive_letter = f"{drive}:"
            ntfs_overlay.mount(drive_letter, fs)
            remap_overlay.map_file(f"{drive_letter}/$mft", fh_mft, f"{drive_letter}/$mft")
            if fh_usnjrnl:
                remap_overlay.map_file(
                    f"{drive_letter}/$extend/$usnjrnl:$J",
                    fh_usnjrnl,
                    f"{drive_letter}/$extend/$usnjrnl:$J",
                )



from __future__ import annotations

import re
from datetime import datetime, timezone
from typing import TYPE_CHECKING

from dissect.target.exceptions import UnsupportedPluginError
from dissect.target.helpers.record import TargetRecordDescriptor
from dissect.target.plugin import Plugin, export

if TYPE_CHECKING:
    from collections.abc import Iterator

AgentExecutorLogRecord = TargetRecordDescriptor(
    "windows/intune/agentexecutor/log",
    [
        ("datetime", "ts"),
        ("string", "component"),
        ("string", "thread"),
        ("string", "type"),
        ("string", "context"),
        ("string", "message"),
        ("string", "file_origin"),
    ],
)

LOG_PATTERN = re.compile(
    r"""
        <!\[LOG\[(?P<message>.*?)\]LOG\]!>
        <time="(?P<hms>\d{2}:\d{2}:\d{2})
            (?:\.(?P<fractional_seconds>\d{1,7}))?"
        \s+
        date="(?P<date>[\d-]+)"
        \s+
        component="(?P<component>[^"]+)"
        \s+
        context="(?P<context>[^"]*)"
        \s+
        type="(?P<type>\d+)"
        \s+
        thread="(?P<thread>\d+)"
        (?:
            \s+file="(?P<file_origin>[^"]*)"
        )?
    """,
    re.DOTALL | re.IGNORECASE | re.VERBOSE,
)


class AgentExecutorLogPlugin(Plugin):
    """Parse Microsoft Intune AgentExecutor logs.

    The AgentExecutor log file captures script execution and system management
    activity from the Microsoft Intune Management Extension agent.
    This plugin parses structured entries and converts them into records suitable
    for timeline and forensic analysis.
    """

    DEFAULT_LOG_PATH = "sysvol/ProgramData/Microsoft/IntuneManagementExtension/Logs/AgentExecutor.log"

    def check_compatible(self) -> None:
        """Verify that the AgentExecutor log file exists within the target.

        Raises:
            UnsupportedPluginError: If the expected log file does not exist.
        """
        if not self.target.fs.path(self.DEFAULT_LOG_PATH).exists():
            raise UnsupportedPluginError(f"AgentExecutor.log not found at {self.DEFAULT_LOG_PATH}")

    @export(record=AgentExecutorLogRecord)
    def agentexecutor(self) -> Iterator[AgentExecutorLogRecord]:
        """Parse the AgentExecutor.log and yield structured records.

        Extracts timestamp, message, context, thread, and type information from
        the AgentExecutor log and yields normalized structured records.

        Yields:
            AgentExecutorLogRecord: A structured representation of each log entry.
        """
        log_path = self.target.fs.path(self.DEFAULT_LOG_PATH)

        try:
            content = log_path.read_text()
        except Exception:
            self.target.log.exception("Failed to open log file %s", log_path)
            return

        match_count = 0

        for match in LOG_PATTERN.finditer(content):
            match_count += 1

            msg = match.group("message").replace("\r", "").strip()
            date_str = match.group("date")
            hms_str = match.group("hms")
            fractional_seconds_str = match.group("fractional_seconds")

            micro_str = (fractional_seconds_str or "000000")[:6].ljust(6, "0")
            time_str = f"{hms_str}.{micro_str}"

            timestamp = None
            for fmt in ("%m-%d-%Y %H:%M:%S.%f", "%d-%m-%Y %H:%M:%S.%f"):
                try:
                    timestamp = datetime.strptime(f"{date_str} {time_str}", fmt).replace(tzinfo=timezone.utc)
                    break
                except ValueError:
                    continue

            if not timestamp:
                self.target.log.debug("Could not parse datetime from %s %s", date_str, time_str)
                continue

            yield AgentExecutorLogRecord(
                ts=timestamp,
                component=match.group("component"),
                thread=match.group("thread"),
                type=match.group("type"),
                context=match.group("context"),
                message=msg,
                file_origin=match.group("file_origin") or "AgentExecutor.log",
                _target=self.target,
            )

        if match_count == 0:
            self.target.log.warning("No log entries matched the regex in %s", log_path)



from __future__ import annotations

from datetime import datetime, timezone
from typing import TYPE_CHECKING

from dissect.target.helpers.regutil import VirtualHive, VirtualKey
from dissect.target.plugins.os.windows.regf.applications import (
    WindowsApplicationsPlugin,
)

if TYPE_CHECKING:
    from dissect.target.target import Target


def test_windows_applications(target_win_users: Target, hive_hklm: VirtualHive) -> None:
    """Test if windows applications are detected correctly in the registry."""

    firefox_name = "SOFTWARE\\Microsoft\\Windows\\CurrentVersion\\Uninstall\\Mozilla Firefox 123.0.1 (x64 nl)"
    firefox_key = VirtualKey(hive_hklm, firefox_name)
    firefox_key.add_value("Comments", "Mozilla Firefox 123.0.1 (x64 nl)")
    firefox_key.add_value("DisplayIcon", "C:\\Program Files\\Mozilla Firefox\\firefox.exe,0")
    firefox_key.add_value("DisplayName", "Mozilla Firefox (x64 nl)")
    firefox_key.add_value("DisplayVersion", "123.0.1")
    firefox_key.add_value("EstimatedSize", 238271)
    firefox_key.add_value("HelpLink", "https://support.mozilla.org")
    firefox_key.add_value("InstallLocation", "C:\\Program Files\\Mozilla Firefox")
    firefox_key.add_value("NoModify", 1)
    firefox_key.add_value("NoRepair", 1)
    firefox_key.add_value("Publisher", "Mozilla")
    firefox_key.add_value("URLInfoAbout", "https://www.mozilla.org")
    firefox_key.add_value("URLUpdateInfo", "https://www.mozilla.org/firefox/123.0.1/releasenotes")
    firefox_key.add_value("UninstallString", '"C:\\Program Files\\Mozilla Firefox\\uninstall\\helper.exe"')
    hive_hklm.map_key(firefox_name, firefox_key)

    chrome_name = "SOFTWARE\\Microsoft\\Windows\\CurrentVersion\\Uninstall\\{47FB91DD-98F3-3C87-A963-357B14EAC7C9}"
    chrome_key = VirtualKey(hive_hklm, chrome_name)
    chrome_key.add_value("DisplayVersion", "122.0.6261.95")
    chrome_key.add_value("InstallDate", "20240301")
    chrome_key.add_value("InstallLocation", "")
    chrome_key.add_value("InstallSource", "C:\\Users\\user\\Desktop\\GoogleChromeEnterpriseBundle64\\Installers\\")
    chrome_key.add_value("ModifyPath", "MsiExec.exe /X{47FB91DD-98F3-3C87-A963-357B14EAC7C9}")
    chrome_key.add_value("NoModify", 1)
    chrome_key.add_value("Publisher", "Google LLC")
    chrome_key.add_value("EstimatedSize", 113725)
    chrome_key.add_value("UninstallString", "MsiExec.exe /X{47FB91DD-98F3-3C87-A963-357B14EAC7C9}")
    chrome_key.add_value("VersionMajor", 70)
    chrome_key.add_value("VersionMinor", 29)
    chrome_key.add_value("WindowsInstaller", 1)
    chrome_key.add_value("Version", 1176322143)
    chrome_key.add_value("Language", 1033)
    chrome_key.add_value("DisplayName", "Google Chrome")
    hive_hklm.map_key(chrome_name, chrome_key)

    addressbook_name = "SOFTWARE\\Microsoft\\Windows\\CurrentVersion\\Uninstall\\AddressBook"
    addressbook_key = VirtualKey(hive_hklm, addressbook_name)
    addressbook_key.timestamp = datetime(2024, 12, 31, 13, 37, 0, tzinfo=timezone.utc)
    hive_hklm.map_key(addressbook_name, addressbook_key)

    msvc_name = "SOFTWARE\\Microsoft\\Windows\\CurrentVersion\\Uninstall\\{D5D19E2F-7189-42FE-8103-92CD1FA457C2}"
    msvc_key = VirtualKey(hive_hklm, msvc_name)
    msvc_key.add_value("DisplayName", "Microsoft Visual C++ 2022 X64 Minimum Runtime - 14.36.32532")
    msvc_key.add_value("InstallDate", "20240301")
    msvc_key.add_value("DisplayVersion", "14.36.32532")
    msvc_key.add_value("Publisher", "Microsoft Corporation")
    msvc_key.add_value(
        "InstallSource",
        "C:\\ProgramData\\Package Cache\\{D5D19E2F-7189-42FE-8103-92CD1FA457C2}v14.36.32532\\packages\\vcRuntimeMinimum_amd64\\",  # noqa: E501
    )
    msvc_key.add_value("SystemComponent", 1)
    hive_hklm.map_key(msvc_name, msvc_key)

    target_win_users.add_plugin(WindowsApplicationsPlugin)
    results = sorted(target_win_users.applications(), key=lambda r: r.name)

    assert len(results) == 4

    assert results[0].ts_installed is None
    assert results[0].ts_modified == datetime(2024, 12, 31, 13, 37, 0, tzinfo=timezone.utc)
    assert results[0].name == "AddressBook"
    assert results[0].type == "system"

    assert results[1].ts_installed == datetime(2024, 3, 1, 0, 0, 0, tzinfo=timezone.utc)
    assert results[1].name == "Google Chrome"
    assert results[1].version == "122.0.6261.95"
    assert results[1].author == "Google LLC"
    assert results[1].type == "user"
    assert results[1].path == "C:\\Users\\user\\Desktop\\GoogleChromeEnterpriseBundle64\\Installers\\"

    assert results[2].ts_installed == datetime(2024, 3, 1, 0, 0, 0, tzinfo=timezone.utc)
    assert results[2].name == "Microsoft Visual C++ 2022 X64 Minimum Runtime - 14.36.32532"
    assert results[2].version == "14.36.32532"
    assert results[2].author == "Microsoft Corporation"
    assert results[2].type == "system"
    assert (
        results[2].path
        == "C:\\ProgramData\\Package Cache\\{D5D19E2F-7189-42FE-8103-92CD1FA457C2}v14.36.32532\\packages\\vcRuntimeMinimum_amd64\\"  # noqa: E501
    )

    assert results[3].ts_installed is None
    assert results[3].name == "Mozilla Firefox (x64 nl)"
    assert results[3].version == "123.0.1"
    assert results[3].author == "Mozilla"
    assert results[3].type == "user"
    assert results[3].path == "C:\\Program Files\\Mozilla Firefox\\firefox.exe,0"



from __future__ import annotations

import bz2
import gzip
import io
import os
import sys
import tempfile
from contextlib import contextmanager
from pathlib import Path
from typing import TYPE_CHECKING, Any
from unittest.mock import Mock, patch

import pytest

from dissect.target.exceptions import (
    FileNotFoundError,
    NotADirectoryError,
    NotASymlinkError,
    SymlinkRecursionError,
)
from dissect.target.filesystem import VirtualFile, VirtualFilesystem
from dissect.target.filesystems.dir import DirectoryFilesystem
from dissect.target.filesystems.ntfs import NtfsFilesystemEntry
from dissect.target.helpers import fsutil

if TYPE_CHECKING:
    from collections.abc import Callable, Iterator

    from dissect.target.target import Target


@pytest.mark.parametrize(
    ("path", "alt_separator", "result"),
    [
        ("/some/dir/some/file", "", "/some/dir/some/file"),
        ("/some/dir/some/file", "\\", "/some/dir/some/file"),
        ("\\some\\dir\\some\\file", "\\", "/some/dir/some/file"),
        ("/some///long\\\\dir/so\\//me\\file", "", "/some/long\\\\dir/so\\/me\\file"),
        ("/some///long\\\\dir/so\\//me\\file", "\\", "/some/long/dir/so/me/file"),
    ],
)
def test_normalize(path: str, alt_separator: str, result: str) -> None:
    assert fsutil.normalize(path, alt_separator=alt_separator) == result


@pytest.mark.parametrize(
    ("args", "alt_separator", "result"),
    [
        (("/some/dir", "some/file"), "", "/some/dir/some/file"),
        (("/some/dir", "some/file"), "\\", "/some/dir/some/file"),
        (("\\some\\dir", "some\\file"), "\\", "/some/dir/some/file"),
        (("/some///long\\\\dir", "so\\//me\\file"), "", "/some/long\\\\dir/so\\/me\\file"),
        (("/some///long\\\\dir", "so\\//me\\file"), "\\", "/some/long/dir/so/me/file"),
    ],
)
def test_join(args: str, alt_separator: str, result: str) -> None:
    assert fsutil.join(*args, alt_separator=alt_separator) == result


@pytest.mark.parametrize(
    ("path", "alt_separator", "result"),
    [
        ("/some/dir/some/file", "", "/some/dir/some"),
        ("/some/dir/some/file", "\\", "/some/dir/some"),
        ("\\some\\dir\\some\\file", "\\", "/some/dir/some"),
        ("/some///long\\\\dir/so\\//me\\file", "", "/some/long\\\\dir/so\\"),
        ("/some///long\\\\dir/so\\//me\\file", "\\", "/some/long/dir/so/me"),
    ],
)
def test_dirname(path: str, alt_separator: str, result: str) -> None:
    assert fsutil.dirname(path, alt_separator=alt_separator) == result


@pytest.mark.parametrize(
    ("path", "alt_separator", "result"),
    [
        ("/some/dir/some/file", "", "file"),
        ("/some/dir/some/file", "\\", "file"),
        ("\\some\\dir\\some\\file", "\\", "file"),
        ("/some///long\\\\dir/so\\//me\\file", "", "me\\file"),
        ("/some///long\\\\dir/so\\//me\\file", "\\", "file"),
    ],
)
def test_basename(path: str, alt_separator: str, result: str) -> None:
    assert fsutil.basename(path, alt_separator=alt_separator) == result


@pytest.mark.parametrize(
    ("path", "alt_separator", "result"),
    [
        ("/some/dir/some/file", "", ("/some/dir/some", "file")),
        ("/some/dir/some/file", "\\", ("/some/dir/some", "file")),
        ("\\some\\dir\\some\\file", "\\", ("/some/dir/some", "file")),
        ("/some/dir/some/", "", ("/some/dir/some", "")),
        ("/some/dir/some\\", "", ("/some/dir", "some\\")),
        ("/some/dir/some/", "\\", ("/some/dir/some", "")),
        ("\\some\\dir\\some\\", "\\", ("/some/dir/some", "")),
        ("/some///long\\\\dir/so\\//me\\file", "", ("/some/long\\\\dir/so\\", "me\\file")),
        ("/some///long\\\\dir/so\\//me\\file", "\\", ("/some/long/dir/so/me", "file")),
    ],
)
def test_split(path: str, alt_separator: str, result: str) -> None:
    assert fsutil.split(path, alt_separator=alt_separator) == result


@pytest.mark.parametrize(
    ("path", "alt_separator", "result"),
    [
        ("/some/dir", "", True),
        ("some/dir", "", False),
        ("\\some/dir", "", False),
        ("/some/dir", "\\", True),
        ("some/dir", "\\", False),
        ("\\some/dir", "\\", True),
    ],
)
def test_isabs(path: str, alt_separator: str, result: str) -> None:
    assert fsutil.isabs(path, alt_separator=alt_separator) == result


@pytest.mark.parametrize(
    ("path", "alt_separator", "result"),
    [
        ("/some/dir/../some/file", "", "/some/some/file"),
        ("/some/dir/../some/file", "\\", "/some/some/file"),
        ("\\some\\dir\\..\\some\\file", "\\", "/some/some/file"),
        ("/some///long\\..\\dir/so\\.//me\\file", "", "/some/long\\..\\dir/so\\./me\\file"),
        ("/some///long\\..\\dir/so\\.//me\\file", "\\", "/some/dir/so/me/file"),
    ],
)
def test_normpath(path: str, alt_separator: str, result: str) -> None:
    assert fsutil.normpath(path, alt_separator=alt_separator) == result


@pytest.mark.parametrize(
    ("path", "cwd", "alt_separator", "result"),
    [
        ("/some/dir", "", "", "/some/dir"),
        ("some/dir", "", "", "/some/dir"),
        ("\\some/dir", "", "", "/\\some/dir"),
        ("/some/dir", "", "\\", "/some/dir"),
        ("some/dir", "", "\\", "/some/dir"),
        ("\\some\\dir", "", "\\", "/some/dir"),
        ("some\\dir", "", "\\", "/some/dir"),
        ("/some/dir", "/my/cwd/", "", "/some/dir"),
        ("\\some\\dir", "\\my\\cwd\\", "\\", "/some/dir"),
        ("some/dir", "/my/cwd/", "", "/my/cwd/some/dir"),
        ("some\\dir", "/my/cwd/", "\\", "/my/cwd/some/dir"),
        ("some/dir", "/my\\cwd/", "", "/my\\cwd/some/dir"),
        ("some\\dir", "/my\\cwd/", "\\", "/my/cwd/some/dir"),
    ],
)
def test_abspath(path: str, cwd: str, alt_separator: str, result: str) -> None:
    assert fsutil.abspath(path, cwd=cwd, alt_separator=alt_separator) == result


@pytest.mark.parametrize(
    ("path", "start", "alt_separator", "result"),
    [
        ("/some/dir/some/file", "/some/dir", "", "some/file"),
        ("/some/dir/some/file", "/some/dir", "\\", "some/file"),
        ("\\some\\dir\\some\\file", "\\some\\dir", "\\", "some/file"),
        ("/some///long\\\\dir/so\\//me\\file", "/some/long/dir", "", "../../long\\\\dir/so\\/me\\file"),
        ("/some///long\\\\dir/so\\//me\\file", "/some/long\\\\dir", "", "so\\/me\\file"),
        ("/some///long\\\\dir/so\\//me\\file", "/some/long/dir", "\\", "so/me/file"),
        ("/some///long\\\\dir/so\\//me\\file", "/some/long\\\\dir", "\\", "so/me/file"),
    ],
)
def test_relpath(path: str, start: str, alt_separator: str, result: str) -> None:
    assert fsutil.relpath(path, start, alt_separator=alt_separator) == result


@pytest.mark.parametrize(
    ("paths", "alt_separator", "result"),
    [
        (["/some/dir/some/file", "/some/dir/some/other"], "", "/some/dir/some"),
        (["/some/dir/some/file", "/some/dir/some/other"], "\\", "/some/dir/some"),
        (["\\some\\dir\\some\\file", "\\some\\dir\\some\\other"], "\\", "/some/dir/some"),
        (["/some/dir/some/file", "/some/dir/other"], "", "/some/dir"),
        (["/some/dir/some/file", "/some/other"], "", "/some"),
        (["/some/dir/some/file", "/some/other"], "\\", "/some"),
    ],
)
def test_commonpath(paths: list[str], alt_separator: str, result: str) -> None:
    assert fsutil.commonpath(paths, alt_separator=alt_separator) == result


def test_isreserved() -> None:
    assert not fsutil.isreserved("CON")
    assert not fsutil.isreserved("foo")


def test_generate_addr() -> None:
    slash_path = "/some/dir/some/file"
    slash_vfs = VirtualFilesystem(alt_separator="")
    slash_target_path = fsutil.TargetPath(slash_vfs, slash_path)

    backslash_path = "\\some\\dir\\some\\file"
    backslash_vfs = VirtualFilesystem(alt_separator="\\")
    backslash_target_path = fsutil.TargetPath(backslash_vfs, backslash_path)

    assert (
        fsutil.generate_addr(slash_path, "")
        == fsutil.generate_addr(slash_path, "\\")
        == fsutil.generate_addr(backslash_path, "\\")
        == fsutil.generate_addr(slash_target_path, "")
        == fsutil.generate_addr(backslash_target_path, "")
        == fsutil.generate_addr(slash_target_path, "\\")
        == fsutil.generate_addr(backslash_target_path, "\\")
    )

    assert fsutil.generate_addr(slash_path, "") != fsutil.generate_addr(backslash_path, "")


def test_stat_result() -> None:
    with pytest.raises(TypeError):
        fsutil.stat_result([0])

    with pytest.raises(TypeError):
        fsutil.stat_result(list(range(100)))

    # ["st_mode", "st_ino", "st_dev", "st_nlink", "st_uid", "st_gid", "st_size", "_st_atime", "_st_mtime", "_st_ctime"]
    values = [0, 1, 2, 3, 4, 5, 6, 7, 8, 9]

    st = fsutil.stat_result(values)
    assert st[0] == st.st_mode == 0
    assert st[9] == 9
    assert st.st_ctime == 9.0
    assert st.st_ctime_ns == 9000000000
    assert st.st_blksize is None
    assert list(st) == [0, 1, 2, 3, 4, 5, 6, 7, 8, 9]

    # [..., "st_atime", "st_mtime", "st_ctime", "st_atime_ns", "st_mtime_ns", "st_ctime_ns"]
    values += [10, 11, 12, 13, 14, 15]

    st = fsutil.stat_result(values)
    assert st[0] == st.st_mode == 0
    assert st[9] == 9
    assert st.st_ctime == 12.0
    assert st.st_ctime_ns == 15
    assert st.st_blksize is None
    assert list(st) == [0, 1, 2, 3, 4, 5, 6, 7, 8, 9]

    my_stat = Path(__file__).stat()
    st = fsutil.stat_result.copy(my_stat)
    assert st == my_stat


@pytest.fixture
def path_fs() -> VirtualFilesystem:
    vfs = VirtualFilesystem()

    vfs.makedirs("/some/dir")
    vfs.makedirs("/some/dir/nested")
    vfs.symlink("/some/dir/file.txt", "/some/symlink.txt")
    vfs.symlink("nonexistent", "/some/dir/link.txt")
    vfs.symlink("/some/dir/nested", "/some/dirlink")
    vfs.map_file_fh("/some/file.txt", io.BytesIO(b"content"))
    vfs.map_file_fh("/some/dir/file.txt", io.BytesIO())
    vfs.map_file_fh("/some/dir/nested/file.txt", io.BytesIO())

    return vfs


def test_target_path_drive(path_fs: VirtualFilesystem) -> None:
    assert path_fs.path("/some/file.txt").drive == ""


def test_target_path_root(path_fs: VirtualFilesystem) -> None:
    assert path_fs.path("/some/file.txt").root == "/"


def test_target_path_anchor(path_fs: VirtualFilesystem) -> None:
    assert path_fs.path("/some/file.txt").anchor == "/"


def test_target_path_parent(path_fs: VirtualFilesystem) -> None:
    assert path_fs.path("/some/dir/file.txt").parent == path_fs.path("/some/dir")


def test_target_path_parents(path_fs: VirtualFilesystem) -> None:
    path = path_fs.path("/some/dir/file.txt")
    parents = list(path.parents)
    assert parents == [path_fs.path("/some/dir"), path_fs.path("/some"), path_fs.path("/")]
    assert [p.exists() for p in parents]
    assert all(p._fs == path_fs for p in parents)
    assert path.parents[1] == path.parents[-2] == path_fs.path("/some")


def test_target_path_name(path_fs: VirtualFilesystem) -> None:
    path = path_fs.path("/some/file.txt")
    assert path.name == "file.txt"


def test_target_path_suffix(path_fs: VirtualFilesystem) -> None:
    path = path_fs.path("/some/file.txt")
    assert path.suffix == ".txt"


def test_target_path_suffixes(path_fs: VirtualFilesystem) -> None:
    path = path_fs.path("/some/file.tar.gz")
    assert path.suffixes == [".tar", ".gz"]


def test_target_path_stem(path_fs: VirtualFilesystem) -> None:
    path = path_fs.path("/some/file.txt")
    assert path.stem == "file"


def test_target_path_as_posix(path_fs: VirtualFilesystem) -> None:
    path = path_fs.path("/some/file.txt")
    assert path.as_posix() == "/some/file.txt"

    path_fs.alt_separator = "\\"
    path = path_fs.path("\\some\\file.txt")
    assert path.exists()
    assert path.as_posix() == "/some/file.txt"


def test_target_path_as_uri(path_fs: VirtualFilesystem) -> None:
    path = path_fs.path("/some/file.txt")
    assert path.as_uri() == "file:///some/file.txt"

    path_fs.alt_separator = "\\"
    path = path_fs.path("\\some\\file.txt")
    assert path.as_uri() == "file:///some/file.txt"


def test_target_path_is_absolute(path_fs: VirtualFilesystem) -> None:
    assert path_fs.path("/some/file.txt").is_absolute()
    assert not path_fs.path("some/file.txt").is_absolute()


def test_target_path_is_relative_to(path_fs: VirtualFilesystem) -> None:
    assert path_fs.path("/some/dir/file.txt").is_relative_to("/some/dir")
    assert not path_fs.path("/some/dir/file.txt").is_relative_to("/some/other")


@pytest.mark.skipif(sys.version_info >= (3, 13), reason="deprecated on Python 3.13+")
def test_target_path_is_reserved(path_fs: VirtualFilesystem) -> None:
    # We currently do not have any reserved names for TargetPath
    assert not path_fs.path("CON").is_reserved()
    assert not path_fs.path("foo").is_reserved()


def test_target_path_join(path_fs: VirtualFilesystem) -> None:
    assert path_fs.path("/some").joinpath("file.txt") == path_fs.path("/some/file.txt")
    assert path_fs.path("/some") / "file.txt" == path_fs.path("/some/file.txt")


def test_target_path_match(path_fs: VirtualFilesystem) -> None:
    assert path_fs.path("/some/file.txt").match("*.txt")
    assert not path_fs.path("/some/file.txt").match("*.csv")


def test_target_path_relative_to(path_fs: VirtualFilesystem) -> None:
    assert path_fs.path("/some/dir/file.txt").relative_to("/some") == path_fs.path("dir/file.txt")


def test_target_path_with_name(path_fs: VirtualFilesystem) -> None:
    assert path_fs.path("/some/file.txt").with_name("new_file.txt") == path_fs.path("/some/new_file.txt")


def test_target_path_with_stem(path_fs: VirtualFilesystem) -> None:
    assert path_fs.path("/some/file.txt").with_stem("new_file") == path_fs.path("/some/new_file.txt")


def test_target_path_with_suffix(path_fs: VirtualFilesystem) -> None:
    assert path_fs.path("/some/file.txt").with_suffix(".csv") == path_fs.path("/some/file.csv")


def test_target_path_stat(path_fs: VirtualFilesystem) -> None:
    stat_result = path_fs.path("/some/file.txt").stat()
    assert stat_result.st_mode == 0o100000
    assert stat_result.st_dev == id(path_fs)
    assert stat_result.st_nlink == 1

    stat_result = path_fs.path("/some").stat()
    assert stat_result.st_mode == 0o40000
    assert stat_result.st_dev == id(path_fs)
    assert stat_result.st_nlink == 1

    assert path_fs.path("/some/symlink.txt").stat() == path_fs.path("/some/dir/file.txt").stat()


def test_target_path_lstat(path_fs: VirtualFilesystem) -> None:
    stat_result = path_fs.path("/some/symlink.txt").lstat()
    assert stat_result != path_fs.path("/some/dir/file.txt").lstat()
    assert stat_result.st_mode == 0o120000


def test_target_path_exists(path_fs: VirtualFilesystem) -> None:
    assert path_fs.path("/some/file.txt").exists()
    assert not path_fs.path("/some/other.txt").exists()


def test_target_path_glob(path_fs: VirtualFilesystem) -> None:
    assert list(path_fs.path("/some").glob("*.txt")) == [
        path_fs.path("/some/symlink.txt"),
        path_fs.path("/some/file.txt"),
    ]
    assert list(path_fs.path("/some").glob("*.csv")) == []


def test_target_path_rglob(path_fs: VirtualFilesystem) -> None:
    assert list(map(str, path_fs.path("/some").rglob("*.txt"))) == [
        "/some/symlink.txt",
        "/some/file.txt",
        "/some/dir/link.txt",
        "/some/dir/file.txt",
        "/some/dir/nested/file.txt",
    ]
    assert list(path_fs.path("/some").rglob("*.TXT")) == []
    assert list(path_fs.path("/some").rglob("*.csv")) == []
    assert list(map(str, path_fs.path("/").rglob("*.*"))) == [
        "/some/symlink.txt",
        "/some/file.txt",
        "/some/dir/link.txt",
        "/some/dir/file.txt",
        "/some/dir/nested/file.txt",
    ]

    with patch.object(path_fs, "case_sensitive", False):
        assert list(map(str, path_fs.path("/some").rglob("*.TXT"))) == [
            "/some/symlink.txt",
            "/some/file.txt",
            "/some/dir/link.txt",
            "/some/dir/file.txt",
            "/some/dir/nested/file.txt",
        ]


@pytest.mark.skipif(sys.version_info < (3, 12), reason="requires Python 3.12+")
def test_target_path_rglob_case_sensitive(path_fs: VirtualFilesystem) -> None:
    assert list(map(str, path_fs.path("/some").rglob("*.TXT", case_sensitive=False))) == [
        "/some/symlink.txt",
        "/some/file.txt",
        "/some/dir/link.txt",
        "/some/dir/file.txt",
        "/some/dir/nested/file.txt",
    ]


@pytest.mark.skipif(sys.version_info < (3, 13), reason="requires Python 3.13+")
def test_target_path_rglob_recurse_symlinks(path_fs: VirtualFilesystem) -> None:
    assert list(map(str, path_fs.path("/some").rglob("*.txt", recurse_symlinks=True))) == [
        "/some/symlink.txt",
        "/some/file.txt",
        "/some/dirlink/file.txt",
        "/some/dir/link.txt",
        "/some/dir/file.txt",
        "/some/dir/nested/file.txt",
    ]


def test_target_path_is_dir(path_fs: VirtualFilesystem) -> None:
    assert path_fs.path("/some/dir").is_dir()
    assert not path_fs.path("/some/file.txt").is_dir()


def test_target_path_is_file(path_fs: VirtualFilesystem) -> None:
    assert path_fs.path("/some/file.txt").is_file()
    assert not path_fs.path("/some/dir").is_file()


def test_target_path_is_mount(path_fs: VirtualFilesystem) -> None:
    assert not path_fs.path("/some").is_mount()

    mnt_vfs = VirtualFilesystem()
    mnt_vfs.makedirs("/foo")
    path_fs.mount("/mnt", mnt_vfs)

    assert path_fs.path("/mnt").is_mount()


def test_target_path_is_symlink(path_fs: VirtualFilesystem) -> None:
    assert path_fs.path("/some/symlink.txt").is_symlink()
    assert not path_fs.path("/some/file.txt").is_symlink()
    assert not path_fs.path("/some/dir").is_symlink()


def test_target_path_is_junction(path_fs: VirtualFilesystem) -> None:
    assert not path_fs.path("/some").is_junction()

    mock_entry = Mock()
    mock_entry.__class__ = NtfsFilesystemEntry
    mock_entry.entry.is_mount_point.return_value = True

    path_fs.map_file_entry("/junction", mock_entry)
    assert path_fs.path("/junction").is_junction()


def test_target_path_is_socket(path_fs: VirtualFilesystem) -> None:
    assert not path_fs.path("/some/file.txt").is_socket()


def test_target_path_is_fifo(path_fs: VirtualFilesystem) -> None:
    assert not path_fs.path("/some/file.txt").is_fifo()


def test_target_path_is_block_device(path_fs: VirtualFilesystem) -> None:
    assert not path_fs.path("/some/file.txt").is_block_device()


def test_target_path_is_char_device(path_fs: VirtualFilesystem) -> None:
    assert not path_fs.path("/some/file.txt").is_char_device()


def test_target_path_iterdir(path_fs: VirtualFilesystem) -> None:
    assert list(path_fs.path("/some").iterdir()) == [
        path_fs.path("/some/dir"),
        path_fs.path("/some/symlink.txt"),
        path_fs.path("/some/dirlink"),
        path_fs.path("/some/file.txt"),
    ]


def test_target_path_walk(path_fs: VirtualFilesystem) -> None:
    assert list(path_fs.path("/some").walk()) == [
        (path_fs.path("/some"), ["dir"], ["symlink.txt", "dirlink", "file.txt"]),
        (path_fs.path("/some/dir"), ["nested"], ["link.txt", "file.txt"]),
        (path_fs.path("/some/dir/nested"), [], ["file.txt"]),
    ]


def test_target_path_open(path_fs: VirtualFilesystem) -> None:
    assert path_fs.path("/some/file.txt").open("rb").read() == b"content"
    assert path_fs.path("/some/file.txt").open("r").read() == "content"


def test_target_path_read_bytes(path_fs: VirtualFilesystem) -> None:
    assert path_fs.path("/some/file.txt").read_bytes() == b"content"


def test_target_path_read_text(path_fs: VirtualFilesystem) -> None:
    assert path_fs.path("/some/file.txt").read_text() == "content"


def test_target_path_readlink(path_fs: VirtualFilesystem) -> None:
    assert path_fs.path("/some/symlink.txt").readlink() == path_fs.path("/some/dir/file.txt")


def test_target_path_resolve(path_fs: VirtualFilesystem) -> None:
    assert path_fs.path("/some/symlink.txt").resolve() == path_fs.path("/some/dir/file.txt")
    assert path_fs.path("/some/symlink.txt").resolve(strict=True) == path_fs.path("/some/dir/file.txt")
    assert path_fs.path("/some/foo").resolve() == path_fs.path("/some/foo")

    with pytest.raises(FileNotFoundError):
        assert path_fs.path("/some/foo").resolve(strict=True)

    with pytest.raises(FileNotFoundError):
        path_fs.path("/some/dir/link.txt").resolve(strict=True)

    with pytest.raises(NotADirectoryError):
        path_fs.path("/some/file.txt/other").resolve(strict=True)


def test_target_path_samefile(path_fs: VirtualFilesystem) -> None:
    assert path_fs.path("/some/symlink.txt").samefile(path_fs.path("/some/dir/file.txt"))
    assert not path_fs.path("/some/symlink.txt").samefile(path_fs.path("/some/file.txt"))


def test_target_path_errors(path_fs: VirtualFilesystem) -> None:
    # TargetPath sometimes emulates OSErrors to play nicely with pathlib, but other times
    # we raise our own FilesystemError variants. Ensure that all user-facing errors are our own.
    path_fs.symlink("symlink1", "symlink2")
    path_fs.symlink("symlink2", "symlink1")

    with pytest.raises(SymlinkRecursionError) as e:
        path_fs.path("symlink1/symlink2/symlink1").resolve()

    # This should raise from the final stat() call
    assert [tb.name for tb in e.traceback[1:3]] == [
        "resolve",
        "stat",
    ]

    with pytest.raises(SymlinkRecursionError) as e:
        path_fs.path("symlink1/symlink2/symlink1").resolve(strict=True)

    # This should raise from the inner realpath() call
    assert [tb.frame.code.name for tb in e.traceback[1:3]] == [
        "resolve",
        "realpath",
    ]

    with pytest.raises(NotASymlinkError):
        path_fs.path("some/file.txt").readlink()

    with pytest.raises(NotADirectoryError):
        path_fs.path("some/file.txt/dir").stat()


def test_target_path_get(path_fs: VirtualFilesystem) -> None:
    # Ensure that TargetPath.get() works as expected
    p = path_fs.path("/some/file.txt")
    assert not hasattr(p, "_entry")

    entry = p.get()
    assert entry is p._entry

    p = next(path_fs.path("/some").iterdir())
    assert not hasattr(p, "_entry")
    assert hasattr(p, "_direntry")

    entry = p.get()
    assert entry is p._entry

    with (
        patch.object(path_fs, "get", side_effect=Exception("Test exception")),
        pytest.raises(Exception, match="Test exception"),
    ):
        path_fs.path("/some/file.txt").get()


def test_target_path_not_implemented(path_fs: VirtualFilesystem) -> None:
    # TargetPath can't do some things, such as write actions or stuff related to a "current" user or path
    # Ensure all those methods properly error
    with pytest.raises(NotImplementedError):
        assert path_fs.path().cwd()

    with pytest.raises(NotImplementedError):
        assert path_fs.path().home()

    with pytest.raises(NotImplementedError):
        assert path_fs.path().expanduser()

    with pytest.raises(NotImplementedError):
        assert path_fs.path().absolute()

    with pytest.raises(NotImplementedError):
        assert path_fs.path().owner()

    with pytest.raises(NotImplementedError):
        assert path_fs.path().group()

    with pytest.raises(NotImplementedError):
        assert path_fs.path().chmod(0o777)

    with pytest.raises(NotImplementedError):
        assert path_fs.path().lchmod(0o777)

    with pytest.raises(NotImplementedError):
        assert path_fs.path().rename("foo")

    with pytest.raises(NotImplementedError):
        assert path_fs.path().replace("foo")

    with pytest.raises(NotImplementedError):
        assert path_fs.path().symlink_to("foo")

    with pytest.raises(NotImplementedError):
        assert path_fs.path().hardlink_to("foo")

    with pytest.raises(NotImplementedError):
        assert path_fs.path().mkdir()

    with pytest.raises(NotImplementedError):
        assert path_fs.path().rmdir()

    with pytest.raises(NotImplementedError):
        assert path_fs.path().touch()

    with pytest.raises(NotImplementedError):
        assert path_fs.path().unlink()

    with pytest.raises(NotImplementedError):
        assert path_fs.path().write_bytes(b"foo")

    with pytest.raises(NotImplementedError):
        assert path_fs.path().write_text("foo")


@pytest.mark.parametrize(
    ("path", "alt_separator", "result"),
    [
        ("/some/dir/some/file", "", ("/", "some", "dir", "some", "file")),
        ("/some/dir//some/file", "", ("/", "some", "dir", "some", "file")),
        ("/some/dir\\some/file", "", ("/", "some", "dir\\some", "file")),
        ("\\some\\dir\\some\\file", "\\", ("/", "some", "dir", "some", "file")),
        ("/some/dir/some/file", "\\", ("/", "some", "dir", "some", "file")),
        ("\\some\\dir\\\\some\\file", "\\", ("/", "some", "dir", "some", "file")),
        ("\\some\\dir/some\\file", "\\", ("/", "some", "dir", "some", "file")),
    ],
)
def test_pure_dissect_path__from_parts(path: str, alt_separator: str, result: tuple[str]) -> None:
    vfs = VirtualFilesystem(alt_separator=alt_separator)
    pure_dissect_path = fsutil.PureDissectPath(vfs, path)

    assert pure_dissect_path.parts == result


@pytest.mark.parametrize(
    ("alt_separator"),
    ["/", "\\"],
)
@pytest.mark.parametrize(
    ("case_sensitive"),
    [True, False],
)
def test_pure_dissect_path__from_parts_flavour(alt_separator: str, case_sensitive: bool) -> None:
    vfs = VirtualFilesystem(alt_separator=alt_separator, case_sensitive=case_sensitive)
    pure_dissect_path = fsutil.PureDissectPath(vfs, "/some/dir")

    obj = getattr(pure_dissect_path, "parser", None) or pure_dissect_path._flavour
    assert obj.altsep == alt_separator
    assert obj.case_sensitive == case_sensitive


def test_pure_dissect_path__from_parts_no_fs_exception() -> None:
    with pytest.raises(TypeError):
        fsutil.PureDissectPath(Mock(), "/some/dir")


@pytest.mark.parametrize(
    ("file_name", "compressor", "content"),
    [
        ("plain", lambda x: x, b"plain\ncontent"),
        ("comp.gz", gzip.compress, b"gzip\ncontent"),
        ("comp_gz", gzip.compress, b"gzip\ncontent"),
        ("comp.bz2", bz2.compress, b"bz2\ncontent"),
        ("comp_bz2", bz2.compress, b"bz2\ncontent"),
    ],
)
def test_open_decompress(file_name: str, compressor: Callable, content: bytes) -> None:
    vfs = VirtualFilesystem()
    fh = io.BytesIO(compressor(content))
    vfs.map_file_fh(file_name, fh)
    assert fsutil.open_decompress(vfs.path(file_name)).read() == content
    fh.seek(2)
    assert fsutil.open_decompress(fileobj=fh).read() == content


@pytest.mark.parametrize(
    ("compressor", "mock_bool_name"),
    [
        (fsutil.lzma.compress, "HAS_XZ"),
        (fsutil.bz2.compress, "HAS_BZ2"),
        (fsutil.zstd.compress, "HAS_ZSTD"),
    ],
)
def test_open_decompress_missing_module(
    compressor: Callable, mock_bool_name: str, monkeypatch: pytest.MonkeyPatch
) -> None:
    vfs = VirtualFilesystem()
    fh = io.BytesIO(compressor(b"hello world"))
    vfs.map_file_fh("compressed_file", fh)

    with monkeypatch.context() as m:
        m.setattr(f"{fsutil.__name__}.{mock_bool_name}", False)
        with pytest.raises(RuntimeError, match=r".* compression detected, but missing optional python module"):
            fsutil.open_decompress(vfs.path("compressed_file"))


def test_open_decompress_text_modes() -> None:
    vfs = VirtualFilesystem()
    vfs.map_file_fh("test", io.BytesIO(b"zomgbbq"))

    fh = fsutil.open_decompress(vfs.path("test"))
    assert not isinstance(fh, io.TextIOWrapper)

    fh = fsutil.open_decompress(vfs.path("test"), "r")
    assert isinstance(fh, io.TextIOWrapper)
    assert fh.encoding == "UTF-8"
    assert fh.errors == "backslashreplace"

    fh = fsutil.open_decompress(vfs.path("test"), "r", errors=None)
    assert isinstance(fh, io.TextIOWrapper)
    assert fh.encoding == "UTF-8"
    assert fh.errors == "strict"

    fh = fsutil.open_decompress(vfs.path("test"), "r", encoding="ascii")
    assert isinstance(fh, io.TextIOWrapper)
    assert fh.encoding == "ascii"
    assert fh.errors == "backslashreplace"


def test_reverse_readlines() -> None:
    vfs = VirtualFilesystem()

    expected_range_reverse = ["99"] + [f"{i}\n" for i in range(98, -1, -1)]

    vfs.map_file_fh("file_n", io.BytesIO("\n".join(map(str, range(100))).encode()))
    assert list(fsutil.reverse_readlines(vfs.path("file_n").open("rt"))) == expected_range_reverse

    vfs.map_file_fh("file_r", io.BytesIO("\r".join(map(str, range(100))).encode()))
    assert list(fsutil.reverse_readlines(vfs.path("file_r").open("rt"))) == expected_range_reverse

    vfs.map_file_fh("file_rn", io.BytesIO("\r\n".join(map(str, range(100))).encode()))
    assert list(fsutil.reverse_readlines(vfs.path("file_rn").open("rt"))) == expected_range_reverse

    vfs.map_file_fh("file_multi", io.BytesIO("🦊\n🦊🦊\n🦊🦊🦊".encode()))
    assert list(fsutil.reverse_readlines(vfs.path("file_multi").open("rt"))) == ["🦊🦊🦊", "🦊🦊\n", "🦊\n"]

    vfs.map_file_fh("file_multi_long", io.BytesIO((("🦊" * 8000) + ("a" * 200) + "\n🦊🦊\n🦊🦊🦊").encode()))
    assert list(fsutil.reverse_readlines(vfs.path("file_multi_long").open("rt"))) == [
        "🦊🦊🦊",
        "🦊🦊\n",
        ("🦊" * 8000) + ("a" * 200) + "\n",
    ]

    vfs.map_file_fh("file_multi_long_single", io.BytesIO((("🦊" * 8000) + ("a" * 200)).encode()))
    assert list(fsutil.reverse_readlines(vfs.path("file_multi_long_single").open("rt"))) == [
        ("🦊" * 8000) + ("a" * 200)
    ]

    vfs.map_file_fh("empty", io.BytesIO())
    assert list(fsutil.reverse_readlines(vfs.path("empty").open("rt"))) == []

    broken_content = (b"foobar\r\n" * 2) + (b"\xc2broken\r\n") + (b"barfoo\r\n" * 2)
    vfs.map_file_fh("file_multi_broken", io.BytesIO(broken_content))
    with pytest.raises(
        UnicodeDecodeError, match="'UTF-8' codec can't decode bytes in position 0-17: failed to decode line"
    ):
        assert list(fsutil.reverse_readlines(vfs.path("file_multi_broken").open("rt"))) == ["barfoo\n", "barfoo\n"]

    assert list(fsutil.reverse_readlines(vfs.path("file_multi_broken").open("rt", errors="backslashreplace"))) == [
        "barfoo\n",
        "barfoo\n",
        "\\xc2broken\n",
        "foobar\n",
        "foobar\n",
    ]


def test_reverse_read() -> None:
    """Test if we read the bytes of a file in reverse succesfully."""
    fs = VirtualFilesystem()

    fs.map_file_fh("file", io.BytesIO(b"1234567890"))
    assert list(fsutil.reverse_read(fs.path("file").open("rb"), chunk_size=2)) == [b"09", b"87", b"65", b"43", b"21"]

    fs.map_file_fh("large_emoji", io.BytesIO(("🐱" * 10_000).encode()))
    content = list(fsutil.reverse_read(fs.path("large_emoji").open("rb")))
    assert len(content) == 5
    assert len(content[0]) == 1024 * 8
    assert len(content[-1]) == 7232
    assert b"".join(content) == bytes(reversed(("🐱" * 10_000).encode()))


@pytest.fixture
def xattrs() -> dict[str, bytes]:
    return {"some_key": b"some_value"}


@pytest.fixture
def listxattr_spec(xattrs: dict[str, str]) -> dict[str, Any]:
    # listxattr() is only available on Linux
    attr_names = list(xattrs.keys())

    if hasattr(os, "listxattr"):
        spec = {
            "create": False,
            "autospec": True,
            "return_value": attr_names,
        }
    else:
        spec = {
            "create": True,
            "return_value": attr_names,
        }

    return spec


@pytest.fixture
def getxattr_spec(xattrs: dict[str, str]) -> dict[str, Any]:
    # getxattr() is only available on Linux
    attr_name = next(iter(xattrs.keys()))
    attr_value = xattrs.get(attr_name)

    if hasattr(os, "getxattr"):
        spec = {
            "create": False,
            "autospec": True,
            "return_value": attr_value,
        }
    else:
        spec = {
            "create": True,
            "return_value": attr_value,
        }

    return spec


@pytest.mark.parametrize(
    "follow_symlinks",
    [
        True,
        False,
    ],
)
def test_fs_attrs(
    xattrs: dict[str, bytes], listxattr_spec: dict[str, Any], getxattr_spec: dict[str, Any], follow_symlinks: bool
) -> None:
    with patch("os.listxattr", **listxattr_spec) as listxattr, patch("os.getxattr", **getxattr_spec) as getxattr:
        path = "/some/path"
        attr_name = next(iter(xattrs.keys()))

        assert fsutil.fs_attrs(path, follow_symlinks=follow_symlinks) == xattrs
        listxattr.assert_called_with(path, follow_symlinks=follow_symlinks)
        getxattr.assert_called_with(path, attr_name, follow_symlinks=follow_symlinks)


@contextmanager
def no_listxattr() -> Iterator[None]:
    if not hasattr(os, "listxattr"):
        yield
        return
    listxattr = os.listxattr
    try:
        del os.listxattr
        yield
    finally:
        os.listxattr = listxattr


def test_fs_attrs_no_os_listxattr() -> None:
    with no_listxattr():
        assert fsutil.fs_attrs("/some/path") == {}


def test_target_path_checks_dirfs(tmp_path: Path, target_win: Target) -> None:
    with tempfile.NamedTemporaryFile(dir=tmp_path, delete=False) as tf:
        tf.write(b"dummy")
        tf.close()

        tmpfile_name = Path(tf.name).name

        fs = DirectoryFilesystem(path=tmp_path)
        target_win.filesystems.add(fs)
        target_win.fs.mount("Z:\\", fs)
        assert target_win.fs.path(f"Z:\\{tmpfile_name}").is_file()
        assert not target_win.fs.path(f"Z:\\{tmpfile_name}\\some").exists()
        assert not target_win.fs.path(f"Z:\\{tmpfile_name}\\some").is_file()


def test_target_path_checks_mapped_dir(tmp_path: Path, target_win: Target) -> None:
    with tempfile.NamedTemporaryFile(dir=tmp_path, delete=False) as tf:
        tf.write(b"dummy")
        tf.close()

        tmpfile_name = Path(tf.name).name

        target_win.filesystems.entries[0].map_dir("test-dir", tmp_path)
        assert target_win.fs.path("C:\\test-dir\\").is_dir()
        assert not target_win.fs.path("C:\\test-dir\\").is_file()

        assert target_win.fs.path(f"C:\\test-dir\\{tmpfile_name}").is_file()
        assert not target_win.fs.path(f"C:\\test-dir\\{tmpfile_name}\\some").is_file()


def test_target_path_checks_virtual() -> None:
    vfs = VirtualFilesystem()
    vfs.map_file_entry("file", VirtualFile(vfs, "file", None))
    assert not vfs.path("file/test").exists()


def test_target_path_backslash_normalisation(target_win: Target, fs_win: VirtualFilesystem, tmp_path: Path) -> None:
    with tempfile.NamedTemporaryFile(dir=tmp_path, delete=False) as tf:
        tf.write(b"dummy")
        tf.close()

        fs_win.map_dir("windows/system32/", tmp_path)
        fs_win.map_file("windows/system32/somefile.txt", tf.name)

        results = list(target_win.fs.path("/").glob("C:\\windows\\system32\\some*.txt"))
        assert len(results) == 1

        results = list(target_win.fs.path("/").glob("sysvol/windows/system32/some*.txt"))
        assert len(results) == 1


@pytest.fixture
def glob_fs() -> VirtualFilesystem:
    vfs = VirtualFilesystem()
    paths = [
        "foo/bar/bla",
        "moo/bar/bla",
        "bar/bla",
    ]
    files = [
        "file.txt",
        "file.ini",
        "other.txt",
    ]
    special_files = [
        "lololo",
        "system.dat",
        "data.tgz",
    ]

    for idx, path in enumerate(paths):
        vfs.makedirs(path)
        for file in files:
            vfs.map_file_entry(f"/{path}/{file}", VirtualFile(vfs, f"{path}/{file}", None))
        special_file = special_files[idx]
        vfs.map_file_entry(f"/{path}/{special_file}", VirtualFile(vfs, f"{path}/{special_file}", None))

    return vfs


@pytest.mark.parametrize(
    ("start_path", "pattern", "results"),
    [
        ("/", "foo/bar/bla/file.*", ["foo/bar/bla/file.ini", "foo/bar/bla/file.txt"]),
        ("/", "foo/bar/*/file.ini", ["foo/bar/bla/file.ini"]),
        ("/", "foo/bar/*/file.*", ["foo/bar/bla/file.ini", "foo/bar/bla/file.txt"]),
        ("/", "*/bar/bla/file.ini", ["foo/bar/bla/file.ini", "moo/bar/bla/file.ini"]),
        ("/", "*/bar/bla/*.ini", ["foo/bar/bla/file.ini", "moo/bar/bla/file.ini"]),
        ("/foo", "*/bla/file.ini", ["foo/bar/bla/file.ini"]),
        ("/foo", "*/bla/*.ini", ["foo/bar/bla/file.ini"]),
        ("/", "boo/bla/*", []),
    ],
)
def test_glob_ext(glob_fs: VirtualFilesystem, start_path: str, pattern: str, results: list[str]) -> None:
    start_entry = glob_fs.get(start_path)
    entries = fsutil.glob_ext(start_entry, pattern)

    entries = sorted([entry.path for entry in entries])
    assert entries == sorted(results)



from __future__ import annotations

from typing import TYPE_CHECKING

from dissect.hypervisor import vbox

from dissect.target import container
from dissect.target.loader import Loader

if TYPE_CHECKING:
    from pathlib import Path

    from dissect.target.target import Target


class VBoxLoader(Loader):
    """Load Oracle VirtualBox files."""

    def __init__(self, path: Path, **kwargs):
        super().__init__(path, **kwargs)
        with path.open("r") as fh:
            self.vbox = vbox.VBox(fh)

    @staticmethod
    def detect(path: Path) -> bool:
        return path.suffix.lower() == ".vbox"

    def map(self, target: Target) -> None:
        for disk in self.vbox.hardware.disks:
            target.disks.add(container.open(self.base_path.joinpath(disk.location)))



from __future__ import annotations

from typing import TYPE_CHECKING

from dissect.target.plugins.os.unix.linux.modules import ModulePlugin
from tests._utils import absolute_path

if TYPE_CHECKING:
    from dissect.target.filesystem import VirtualFilesystem
    from dissect.target.target import Target


def test_modules_plugin(target_unix: Target, fs_unix: VirtualFilesystem) -> None:
    test_folder = absolute_path("_data/plugins/os/unix/linux/modules/module")
    fs_unix.map_dir("/sys/module", test_folder)

    target_unix.add_plugin(ModulePlugin)
    results = sorted(target_unix.sysmodules(), key=lambda x: x.name)
    assert len(results) == 2
    assert results[0].name == "modulea"
    assert results[0].size == 1
    assert results[0].refcount == 3
    assert results[0].used_by == ["holdera"]
    assert results[1].name == "moduleb"
    assert results[1].size == 2
    assert results[1].refcount == 4
    assert sorted(results[1].used_by) == ["holdera", "holderb"]



from __future__ import annotations

from typing import TYPE_CHECKING, TypeVar

from dissect.target.helpers.nfs.nfs3 import (
    CookieVerf,
    DirOpArgs,
    EntryPlus,
    FileAttributes,
    FileHandle,
    FileType,
    LookupResult,
    MountOK,
    MountStat,
    NfsStat,
    NfsTime,
    ReadDirPlusParams,
    ReadDirPlusResult,
    ReadlinkResult,
    ReadParams,
    ReadResult,
    SpecData,
)
from dissect.target.helpers.sunrpc.serializer import (
    Int32Serializer,
    OpaqueVarLengthSerializer,
    XdrDeserializer,
    XdrSerializer,
)
from dissect.target.helpers.sunrpc.sunrpc import Bool

if TYPE_CHECKING:
    import io


class MountResultDeserializer(XdrDeserializer[MountOK | MountStat]):
    def deserialize(self, payload: io.BytesIO) -> MountOK | MountStat:
        mount_stat = self._read_enum(payload, MountStat)
        if mount_stat != MountStat.OK:
            return mount_stat

        filehandle_bytes = self._read_var_length_opaque(payload)
        auth_flavors = self._read_var_length(payload, Int32Serializer())

        return MountOK(FileHandle(filehandle_bytes), auth_flavors)


class ReadDirPlusParamsSerializer(XdrSerializer[ReadDirPlusParams]):
    def serialize(self, params: ReadDirPlusParams) -> bytes:
        result = self._write_var_length_opaque(params.dir.opaque)
        result += self._write_uint64(params.cookie)
        result += self._write_var_length_opaque(params.cookieverf.opaque)
        result += self._write_uint32(params.dir_count)
        result += self._write_uint32(params.max_count)

        return result


class SpecDataSerializer(XdrDeserializer[SpecData]):
    def deserialize(self, payload: io.BytesIO) -> bytes:
        specdata1 = self._read_uint32(payload)
        specdata2 = self._read_uint32(payload)

        return SpecData(specdata1, specdata2)


class NfsTimeSerializer(XdrDeserializer[NfsTime]):
    def deserialize(self, payload: io.BytesIO) -> bytes:
        seconds = self._read_uint32(payload)
        nseconds = self._read_uint32(payload)

        return NfsTime(seconds, nseconds)


class FileAttributesSerializer(XdrDeserializer[FileAttributes]):
    def deserialize(self, payload: io.BytesIO) -> FileAttributes:
        type = self._read_enum(payload, FileType)
        mode = self._read_uint32(payload)
        nlink = self._read_uint32(payload)
        uid = self._read_uint32(payload)
        gid = self._read_uint32(payload)
        size = self._read_uint64(payload)
        used = self._read_uint64(payload)
        rdev = SpecDataSerializer().deserialize(payload)
        fsid = self._read_uint64(payload)
        fileid = self._read_uint64(payload)
        time_deserializer = NfsTimeSerializer()
        atime = time_deserializer.deserialize(payload)
        mtime = time_deserializer.deserialize(payload)
        ctime = time_deserializer.deserialize(payload)

        return FileAttributes(type, mode, nlink, uid, gid, size, used, rdev, fsid, fileid, atime, mtime, ctime)


class EntryPlusSerializer(XdrDeserializer[EntryPlus]):
    def deserialize(self, payload: io.BytesIO) -> EntryPlus:
        fileid = self._read_uint64(payload)
        name = self._read_string(payload)
        cookie = self._read_uint64(payload)
        attributes = self._read_optional(payload, FileAttributesSerializer())
        handle_bytes = self._read_optional(payload, OpaqueVarLengthSerializer())
        handle = FileHandle(handle_bytes) if handle_bytes is not None else None

        return EntryPlus(fileid, name, cookie, attributes, handle)


class ReadDirPlusResultDeserializer(XdrDeserializer[ReadDirPlusResult]):
    def deserialize(self, payload: io.BytesIO) -> ReadDirPlusResult:
        dir_attributes = self._read_optional(payload, FileAttributesSerializer())
        cookieverf = self._read_var_length_opaque(payload)

        entries = list[EntryPlus]()
        while True:
            entry = self._read_optional(payload, EntryPlusSerializer())
            if entry is None:
                break

            entries.append(entry)

        eof = self._read_enum(payload, Bool) == Bool.TRUE

        return ReadDirPlusResult(dir_attributes, CookieVerf(cookieverf), entries, eof)


class Read3ArgsSerializer(XdrSerializer[ReadDirPlusParams]):
    def serialize(self, args: ReadParams) -> bytes:
        result = self._write_var_length_opaque(args.file.opaque)
        result += self._write_uint64(args.offset)
        result += self._write_uint32(args.count)
        return result


# In contrast to rfc we do not return file attributes on failure
class Read3ResultDeserializer(XdrDeserializer[ReadResult]):
    def deserialize(self, payload: io.BytesIO) -> ReadResult:
        file_attributes = self._read_optional(payload, FileAttributesSerializer())
        count = self._read_uint32(payload)
        eof = self._read_enum(payload, Bool) == Bool.TRUE
        data = self._read_var_length_opaque(payload)
        return ReadResult(file_attributes, count, eof, data)


class DirOpArgs3Serializer(XdrSerializer[DirOpArgs]):
    def serialize(self, args: DirOpArgs) -> bytes:
        result = self._write_var_length_opaque(args.handle.opaque)
        result += self._write_string(args.filename)
        return result


class Lookup3ResultDeserializer(XdrDeserializer[LookupResult]):
    def deserialize(self, payload: io.BytesIO) -> LookupResult:
        handle_bytes = self._read_var_length_opaque(payload)
        attribute_serializer = FileAttributesSerializer()
        object_attributes = self._read_optional(payload, attribute_serializer)
        dir_attributes = self._read_optional(payload, attribute_serializer)

        return LookupResult(
            object=FileHandle(handle_bytes), obj_attributes=object_attributes, dir_attributes=dir_attributes
        )


class ReadLink3ResultDeserializer(XdrDeserializer[ReadlinkResult]):
    def deserialize(self, payload: io.BytesIO) -> ReadlinkResult:
        attributes = self._read_optional(payload, FileAttributesSerializer())
        target = self._read_string(payload)

        return ReadlinkResult(attributes, target)


ResultType = TypeVar("ResultType")


class ResultDeserializer(XdrDeserializer[ResultType | NfsStat]):
    """A higher order deserializer that returns a result or an NFS status."""

    def __init__(self, deserializer: XdrDeserializer[ResultType]):
        self._deserializer = deserializer

    def deserialize(self, payload: io.BytesIO) -> ResultType | NfsStat:
        stat = self._read_enum(payload, NfsStat)
        if stat != NfsStat.OK:
            return stat

        return self._deserializer.deserialize(payload)
